import json
import networkx as nx
import numpy as np
import random
import itertools

from ipycanvas import Canvas
from os.path import join
from natsort import natsorted
from matplotlib import pyplot as plt
import ipywidgets as widgets
from IPython.display import display


class Simulation:
    def __init__(
        self,
        field_save_dir: str,
        t_schedule: callable,
        random_search: callable,
        stochastic_hc: callable,
    ):
        self.path = []
        self.state = []
        self.last_state = []
        self.cost_history = []
        self.best_state = []
        self.best_state_cost = 0
        self.best_path = []
        self._load_field(save_dir=field_save_dir)
        self.neighbor_fig = plt.figure()
        self.neighbor_ax = self.neighbor_fig.add_subplot()
        self.t_schedule = t_schedule
        self.random_search = random_search
        self.stochastic_hc = stochastic_hc

    def _load_field(self, save_dir: str):
        with open(join(save_dir, "node_link_data.json"), "r") as f:
            node_link_data = json.load(f)
        self.sim_graph: nx.DiGraph = nx.node_link_graph(node_link_data)

        with open(join(save_dir, "field_info.json"), "r") as f:
            self.field_info_dict = json.load(f)

        with open(join(save_dir, "field.json"), "r") as f:
            field_dict = json.load(f)

        self._prepare_field_for_sim(field_dict)

    def _prepare_field_for_sim(self, field_dict: dict):
        # get border from field dict
        self.field_border = field_dict["field_border"]
        # save logistic point names
        n_logistic_points = 1
        # n_logistic_points = self.logistic_point_attributes["quantity"]
        self.logistic_point_names = []
        for i in range(n_logistic_points):
            self.logistic_point_names.append(f"logistic{i}")

        # Create sub_graph for navigation only on headlands
        all_edges_list = list(self.sim_graph.edges)
        headland_edges = [
            e for e in all_edges_list if self.sim_graph.edges()[e]["headland"] == True
        ]
        self.headland_subgraph = self.sim_graph.edge_subgraph(headland_edges)
        # init global info of free rows
        self.not_seeded_and_targeted_rows = [
            e
            for e in all_edges_list
            if not self.sim_graph.edges()[e]["seeded"]
            and not self.sim_graph.edges()[e]["headland"] == True
            and not self.sim_graph.edges()[e]["targeted"]
        ]

        # simplify graph
        simple_graph = nx.DiGraph()
        for node in list(self.sim_graph.nodes()):
            for edge in self.sim_graph.edges(node):
                simple_edge = (
                    self.simplify_node_name(edge[0]),
                    self.simplify_node_name(edge[1]),
                )
                if simple_edge in list(simple_graph.edges()):
                    continue
                simple_graph.add_node(
                    simple_edge[0], pos=self.sim_graph.nodes()[node]["pos"]
                )
                path_x = self.sim_graph.edges()[edge]["path_x"]
                path_y = self.sim_graph.edges()[edge]["path_y"]
                dist = self.sim_graph.edges()[edge]["dist"]
                headland = self.sim_graph.edges()[edge]["headland"]
                simple_graph.add_edge(
                    *simple_edge,
                    dist=dist,
                    headland=headland,
                    path_x=path_x,
                    path_y=path_y,
                )
        self.sim_graph = simple_graph

    def prepare_complete_state_representation(self):
        # create headland subgraph
        headland_edges = [
            e
            for e in list(self.sim_graph.edges())
            if self.sim_graph.edges[e]["headland"]
        ]
        self.headland_subgraph = nx.edge_subgraph(self.sim_graph, headland_edges)

        # create solution representation
        self.rows = []
        for edge in list(self.sim_graph.edges()):
            if not self.sim_graph.edges()[edge]["headland"]:
                if not int(edge[0]) % 2 == 0:
                    self.rows.append((edge, (edge[1], edge[0])))

    def create_random_state(self) -> list[str, str]:
        state = []
        random.shuffle(self.rows)
        for row in self.rows:
            state.append(random.choice(row))
        return state

    def find_closest_path(
        self, origin: str, edge: tuple[str, str]
    ) -> tuple[tuple[str, str], float, list]:
        dist_to_row_start_0 = nx.shortest_path_length(
            G=self.headland_subgraph,
            source=origin,
            target=edge[0],
            weight="dist",
        )
        dist_to_row_start_1 = nx.shortest_path_length(
            G=self.headland_subgraph,
            source=origin,
            target=edge[1],
            weight="dist",
        )
        if dist_to_row_start_0 >= dist_to_row_start_1:
            edge = (edge[1], edge[0])
            dist = dist_to_row_start_1
        else:
            dist = dist_to_row_start_0
        path = nx.shortest_path(
            G=self.headland_subgraph,
            source=origin,
            target=edge[0],
            weight="dist",
        )
        return edge, dist, path

    def evaluate_state(self, state: list = None) -> float:
        if state is None:
            state = self.state
        headland_dist = 0.0
        last_node = ""
        self.path = []
        for idx, row in enumerate(state):
            if idx == 0:
                start_node = self.logistic_point_names[0]
            else:
                start_node = last_node

            edge, dist_to_row, path = self.find_closest_path(
                origin=start_node, edge=row
            )

            headland_dist += dist_to_row
            self.path.extend(path)
            last_node = edge[1]

        dist_to_logistic = nx.shortest_path_length(
            G=self.headland_subgraph,
            source=state[-1][1],
            target=self.logistic_point_names[0],
            weight="dist",
        )
        self.path.extend(
            nx.shortest_path(
                G=self.headland_subgraph,
                source=last_node,
                target=self.logistic_point_names[0],
                weight="dist",
            )
        )
        headland_dist += dist_to_logistic
        return headland_dist

    def init_optimization(self, optimize_method: str, keep_history: bool = False):
        self.last_state = []
        self.optimize_method = optimize_method
        self.state = self.create_random_state()
        cost = self.evaluate_state()
        if not keep_history:
            self.best_path = self.path
            self.best_state_cost = cost
            self.best_state = self.state
            self.cost_history = []
            self.cost_history.append(round(cost, 2))

    def create_neighborhood(self):
        self.last_state = self.state
        combinations = list(itertools.combinations(range(0, len(self.state)), 2))
        unique_combinations = list(set(combinations))
        random.shuffle(unique_combinations)
        self.draw_neighborhood(neighborhood=unique_combinations)
        self.neighborhood = unique_combinations

    def draw_neighborhood(self, neighborhood: list):
        plt.cla()
        costs = []
        for neighbor in neighborhood:
            neighbor_state = self.apply_neighbor_to_state(neighbor)
            cost = self.evaluate_state(state=neighbor_state)
            costs.append(cost)
        costs = np.array(costs)

        self.neighbor_ax.scatter(0, self.evaluate_state(self.state), c="tab:red")
        for idx, cost in zip(range(len(neighborhood)), reversed(costs)):
            self.neighbor_ax.bar(idx + 1, cost, color="tab:blue")
        # min line
        self.neighbor_ax.plot(
            [0, len(neighborhood) + 1],
            [costs.min(), costs.min()],
            "b:",
        )
        # current line
        current_state_cost = self.evaluate_state(self.state)
        self.neighbor_ax.plot(
            [0, len(neighborhood) + 1],
            [current_state_cost, current_state_cost],
            "r:",
        )
        # plot cost history
        cost_plot_idcs = [
            i + len(neighborhood) + 1 for i in range(len(self.cost_history))
        ]
        self.neighbor_ax.plot(cost_plot_idcs, self.cost_history)

        # x-axis labels
        neighborhood_labels = [
            (self.state[n[0]], self.state[n[1]]) for n in reversed(neighborhood)
        ]
        self.neighbor_ax.set_xticks(
            list(range(len(neighborhood) + 1 + len(self.cost_history))),
            ["current", *neighborhood_labels, *list(range(len(self.cost_history)))],
            rotation=90,
        )

        self.neighbor_fig.suptitle("Neighborhood & cost history")
        self.neighbor_ax.set_ylabel("Cost")
        self.neighbor_fig.subplots_adjust(bottom=0.35)

    def apply_neighbor_to_state(self, neighbor: list[int, int]) -> list:
        state = self.state.copy()
        idx1, idx2 = neighbor
        old_val_idx1 = state[idx1]
        state[idx1] = state[idx2]
        state[idx2] = old_val_idx1

        return state

    def hc_successor(self) -> [list, bool]:
        terminate = True
        cost = self.evaluate_state()
        selected_neighbor_state = self.state
        for neighbor in self.neighborhood:
            neighbor_state = self.apply_neighbor_to_state(neighbor=neighbor)
            neighbor_cost = self.evaluate_state(state=neighbor_state)
            if neighbor_cost < cost:
                cost = neighbor_cost
                selected_neighbor_state = neighbor_state
                terminate = False

        return selected_neighbor_state, terminate

    def first_choice_hc_successor(self) -> [list, bool]:
        cost = self.cost_history[-1]
        local_neighborhood = self.neighborhood.copy()
        while cost >= self.cost_history[-1]:
            neighbor = local_neighborhood.pop()
            neighbor_state = self.apply_neighbor_to_state(neighbor=neighbor)
            cost = self.evaluate_state(state=neighbor_state)
            if not local_neighborhood:
                return self.state, True

        return neighbor_state, False

    def stochastic_hc_successor(self) -> [list, bool]:
        neighbor_states = []
        neighbor_costs = []
        for neighbor in self.neighborhood:
            neighbor_state = self.apply_neighbor_to_state(neighbor=neighbor)
            neighbor_cost = self.evaluate_state(state=neighbor_state)
            neighbor_costs.append(neighbor_cost)
            neighbor_states.append(neighbor_state)
        return self.stochastic_hc(
            current_state=self.state,
            neighbor_states=neighbor_states,
            neighbor_costs=neighbor_costs,
        )

    def random_restart_hc_successor(self) -> [list, bool]:
        hc_successor, terminate = self.hc_successor()
        if terminate:
            # perform restart
            self.init_optimization(
                optimize_method=self.optimize_method, keep_history=True
            )
            hc_successor = self.state
        return hc_successor, False

    def sa_successor(self) -> [list, bool]:
        # get t
        t = len(self.cost_history)
        # calculate temperature with function of user
        temperature = self.t_schedule(t) * 100
        if temperature <= 0:
            return self.state, True
        # choose random successor
        local_neighborhood = self.neighborhood.copy()
        neighbor_candidate = local_neighborhood.pop()
        neighbor_candidate_state = self.apply_neighbor_to_state(
            neighbor=neighbor_candidate
        )
        # get cost of the successor
        cost = self.evaluate_state(state=neighbor_candidate_state)

        delta_e = cost - self.cost_history[-1]

        if delta_e < 0:
            return neighbor_candidate_state, False
        else:
            propability = np.exp(-delta_e / temperature)
            print(
                f"Keine Verbesserung mit delta_e {delta_e:.2f}, temperature {temperature / 100:.2f} und Annahmewahrscheinlichkeit {propability * 100:.2f}%"
            )
            if random.random() <= propability:
                return neighbor_candidate_state, False
            else:
                return self.state, False

    def random_search_successor(self) -> [list, bool]:
        neighbor_states = []
        neighbor_costs = []
        for neighbor in self.neighborhood:
            neighbor_state = self.apply_neighbor_to_state(neighbor=neighbor)
            neighbor_cost = self.evaluate_state(state=neighbor_state)
            neighbor_costs.append(neighbor_cost)
            neighbor_states.append(neighbor_state)
        return self.random_search(
            current_state=self.state,
            neighbor_states=neighbor_states,
            neighbor_costs=neighbor_costs,
        )

    def set_successor(self):
        # method
        if self.optimize_method == "Hill climbing":
            chosen_state, terminate = self.hc_successor()
        elif self.optimize_method == "First-choice HC":
            chosen_state, terminate = self.first_choice_hc_successor()
        elif self.optimize_method == "Stochastic HC":
            chosen_state, terminate = self.stochastic_hc_successor()
        elif self.optimize_method == "Random-restart HC":
            chosen_state, terminate = self.random_restart_hc_successor()
        elif self.optimize_method == "SA":
            chosen_state, terminate = self.sa_successor()
        elif self.optimize_method == "Random search":
            chosen_state, terminate = self.random_search_successor()
        else:
            return

        self.state = chosen_state

        cost = self.evaluate_state()
        if cost < self.best_state_cost:
            self.best_state_cost = cost
            self.best_state = self.state
            self.best_path = self.path
        if not terminate:
            self.cost_history.append(round(cost, 2))
        return terminate

    @staticmethod
    def simplify_node_name(node: str) -> str:
        node = node.replace("anticlock_", "")
        node = node.replace("clock_", "")
        node = node.replace("_in", "")
        node = node.replace("_out", "")
        return node


class FieldVisualizer:
    def __init__(self, simulation: Simulation, canvas_size: int = 400):
        self.sim_canvas_size = canvas_size
        self.simulation_canvas = Canvas(
            height=self.sim_canvas_size, width=int(self.sim_canvas_size * 16 / 9)
        )
        self._calculate_scaling_params_for_simulation(simulation=simulation)

    def _calculate_scaling_params_for_simulation(self, simulation: Simulation):
        field_border_x = np.array(simulation.field_border[0])
        field_border_y = np.array(simulation.field_border[1])

        min_x = field_border_x.min()
        max_x = field_border_x.max()
        min_y = field_border_y.min()
        max_y = field_border_y.max()

        delta_x = max_x - min_x
        delta_y = max_y - min_y

        if delta_x > delta_y:
            high_delta = delta_x
        else:
            high_delta = delta_y

        self.sim_canvas_factor = self.sim_canvas_size / high_delta
        self.sim_canvas_x_offset = (
            abs(min_x) + (self.sim_canvas_size / self.sim_canvas_factor - delta_x) / 2
        )
        self.sim_canvas_y_offset = (
            abs(min_y) + (self.sim_canvas_size / self.sim_canvas_factor - delta_y) / 2
        )

        scaled_field_border_x = (
            field_border_x + self.sim_canvas_x_offset
        ) * self.sim_canvas_factor
        scaled_field_border_y = (
            field_border_y + self.sim_canvas_y_offset
        ) * self.sim_canvas_factor
        self.scaled_field_border = [
            (fb_x, fby)
            for fb_x, fby in zip(scaled_field_border_x, scaled_field_border_y)
        ]

        self.rows = [
            e
            for e in simulation.sim_graph.edges()
            if simulation.sim_graph.edges[e]["headland"] == False
        ]

        x_row_coords = [
            np.array(
                (
                    simulation.sim_graph.nodes[a]["pos"][0],
                    simulation.sim_graph.nodes[b]["pos"][0],
                )
            )
            for a, b in self.rows
        ]
        y_row_coords = [
            np.array(
                (
                    simulation.sim_graph.nodes[a]["pos"][1],
                    simulation.sim_graph.nodes[b]["pos"][1],
                )
            )
            for a, b in self.rows
        ]
        row_coords = np.array((x_row_coords, y_row_coords))
        scaled_row_coords = self._rescale_to_simulation_frame_coords(row_coords)

        self.scaled_row_coords = [
            ([x[0], y[0]], [x[1], y[1]])
            for x, y in zip(scaled_row_coords[0], scaled_row_coords[1])
        ]

        logistic_nodes = []
        for names in simulation.logistic_point_names:
            logistic_nodes.append(names)

        self.scaled_logistic_coords = []
        for logistic_node in logistic_nodes:
            logistic_pos = np.array(simulation.sim_graph.nodes[logistic_node]["pos"])
            logistic_pos = self._rescale_to_simulation_frame_coords(logistic_pos)

    def _rescale_to_simulation_frame_coords(
        self, coord: np.array
    ) -> tuple[np.array, np.array]:
        return (coord[0] + self.sim_canvas_x_offset) * self.sim_canvas_factor, (
            coord[1] + self.sim_canvas_y_offset
        ) * self.sim_canvas_factor

    def draw_borders(self):
        self.simulation_canvas.stroke_style = "#000000"
        self.simulation_canvas.line_width = 1
        self.simulation_canvas.stroke_polygon(self.scaled_field_border)
        self.simulation_canvas.fill_style = "#997950"
        self.simulation_canvas.fill_polygon(self.scaled_field_border)

    def draw_nodes(self, simulation: Simulation):
        radius = 8
        for node in list(simulation.sim_graph.nodes()):
            pos = simulation.sim_graph.nodes()[node]["pos"]
            scaled_pos_x, scaled_pos_y = self._rescale_to_simulation_frame_coords(
                np.array((pos[0], pos[1]))
            )
            # set color
            if "logistic" in node:
                self.simulation_canvas.fill_style = "#800080"
            else:
                self.simulation_canvas.fill_style = "#808080"

            self.simulation_canvas.fill_circle(
                scaled_pos_x,
                scaled_pos_y,
                radius,
            )
            self.simulation_canvas.fill_style = "#000000"
            self.simulation_canvas.fill_text(
                text=node,
                x=scaled_pos_x,
                y=scaled_pos_y,
            )

    def draw_rows(self, simulation: Simulation):
        row_width = (
            simulation.field_info_dict["operating_width"] * self.sim_canvas_factor + 1
        )

        self.simulation_canvas.stroke_style = "#4c392d"
        self.simulation_canvas.line_width = row_width
        for row_coord in self.scaled_row_coords:
            self.simulation_canvas.stroke_line(*row_coord[0], *row_coord[1])

    @staticmethod
    def rgb2hex(r, g, b):
        return "#{:02x}{:02x}{:02x}".format(r, g, b)

    def draw_current_solution_path(
        self, simulation: Simulation, draw_best: bool = False
    ):
        if draw_best:
            path = simulation.best_path
        else:
            path = simulation.path
        self.simulation_canvas.line_width = 3
        for idx, node in enumerate(path):
            r = int(idx / len(path) * 200)
            g = 200 - int(idx / len(path) * 200)
            self.simulation_canvas.stroke_style = self.rgb2hex(r, g, 140)
            if idx == 0:
                continue
            edge = (path[idx - 1], node)
            dubin_x = simulation.sim_graph.edges()[edge]["path_x"]
            dubin_y = simulation.sim_graph.edges()[edge]["path_y"]
            rescaled_path = self._rescale_to_simulation_frame_coords(
                np.array([dubin_x, dubin_y])
            )
            lines = [
                np.array((x, y)) for x, y in zip(rescaled_path[0], rescaled_path[1])
            ]
            self.simulation_canvas.begin_path()
            self.simulation_canvas.move_to(*lines[0])
            for p in lines:
                self.simulation_canvas.line_to(*p)

            self.simulation_canvas.stroke()

    def draw_path_segment(
        self, simulation: Simulation, idx: int, draw_best: bool = False
    ):
        if draw_best:
            path = simulation.best_path
        else:
            path = simulation.path
        self.simulation_canvas.line_width = 3
        if idx == 0:
            return
        node = path[idx]
        r = int(idx / len(path) * 200)
        g = 200 - int(idx / len(path) * 200)
        self.simulation_canvas.stroke_style = self.rgb2hex(r, g, 140)
        edge = (path[idx - 1], node)
        dubin_x = simulation.sim_graph.edges()[edge]["path_x"]
        dubin_y = simulation.sim_graph.edges()[edge]["path_y"]
        rescaled_path = self._rescale_to_simulation_frame_coords(
            np.array([dubin_x, dubin_y])
        )
        lines = [np.array((x, y)) for x, y in zip(rescaled_path[0], rescaled_path[1])]
        self.simulation_canvas.begin_path()
        self.simulation_canvas.move_to(*lines[0])
        for p in lines:
            self.simulation_canvas.line_to(*p)

        self.simulation_canvas.stroke()

    def draw_available_path(self, simulation: Simulation, current_node: str):
        self.simulation_canvas.line_width = 3
        self.simulation_canvas.stroke_style = "green"
        for edge in list(simulation.sim_graph.edges(current_node)):
            dubin_x = simulation.sim_graph.edges()[edge]["path_x"]
            dubin_y = simulation.sim_graph.edges()[edge]["path_y"]
            rescaled_path = self._rescale_to_simulation_frame_coords(
                np.array([dubin_x, dubin_y])
            )
            lines = [
                np.array((x, y)) for x, y in zip(rescaled_path[0], rescaled_path[1])
            ]
            self.simulation_canvas.begin_path()
            self.simulation_canvas.move_to(*lines[0])
            for p in lines:
                self.simulation_canvas.line_to(*p)
            self.simulation_canvas.stroke()

    def draw_states(self, simulation: Simulation):
        # clear info side
        self.simulation_canvas.clear_rect(
            self.sim_canvas_size + 10, 0, self.sim_canvas_size, self.sim_canvas_size
        )
        self.simulation_canvas.fill_style = "#000000"
        self.simulation_canvas.font = "15px Arial"

        self.simulation_canvas.fill_text(
            text="last state",
            x=self.sim_canvas_size + 10,
            y=20,
        )
        self.simulation_canvas.fill_text(
            text="current state",
            x=self.sim_canvas_size + 100,
            y=20,
        )
        self.simulation_canvas.fill_text(
            text="cost history",
            x=self.sim_canvas_size + 200,
            y=20,
        )
        self.simulation_canvas.fill_text(
            text="best",
            x=self.sim_canvas_size + 300,
            y=20,
        )
        # find indices of changed rows
        changed_state_idx = []
        if simulation.last_state:
            changed_state_idx = [
                idx
                for idx, item in enumerate(simulation.state)
                if item != simulation.last_state[idx]
            ]

        for idx, row in enumerate(simulation.last_state):
            self.simulation_canvas.fill_text(
                text=row,
                x=self.sim_canvas_size + 10,
                y=50 + idx * 16,
            )

        for idx, row in enumerate(simulation.state):
            # highlight changed rows
            if idx in changed_state_idx:
                self.simulation_canvas.fill_style = "#9F0000"
            self.simulation_canvas.fill_text(
                text=row,
                x=self.sim_canvas_size + 100,
                y=50 + idx * 16,
            )
            self.simulation_canvas.fill_style = "#000000"

        # draw cost history
        for idx, cost in enumerate(simulation.cost_history):
            self.simulation_canvas.fill_text(
                text=cost,
                x=self.sim_canvas_size + 200,
                y=50 + idx * 16,
            )
        # draw best state
        for idx, row in enumerate(simulation.best_state):
            self.simulation_canvas.fill_text(
                text=row,
                x=self.sim_canvas_size + 300,
                y=50 + idx * 16,
            )
        # best cost
        self.simulation_canvas.fill_text(
            text=f"cost={simulation.best_state_cost:.2f}",
            x=self.sim_canvas_size + 300,
            y=50 + len(simulation.best_state) * 16,
        )

    def draw_optimization_frame(
        self, simulation: Simulation, draw_idx: int = None, draw_best: bool = False
    ):
        if draw_idx is None or draw_idx == 0:
            self.draw_borders()
            self.draw_rows(simulation=simulation)
            self.draw_nodes(simulation=simulation)
            self.draw_states(simulation=simulation)
        if draw_idx is None:
            self.draw_current_solution_path(simulation=simulation, draw_best=draw_best)
        else:
            self.draw_path_segment(
                simulation=simulation, idx=draw_idx, draw_best=draw_best
            )


class IPyUI:
    def __init__(
        self, field_vis: FieldVisualizer, optimize_sim: Simulation, max_iterations: int
    ):
        self.field_vis = field_vis
        self.optimize_sim = optimize_sim
        self.current_idx = 0
        self.showing_best = False
        self.max_iterations = max_iterations
        self.iteration = 0

    def optimization_ui(self):
        # Optimization settings
        self.optimize_method = "Hill climbing"
        self.tpye_dd = widgets.Dropdown(
            options=[
                "Hill climbing",
                "First-choice HC",
                "Stochastic HC",
                "Random-restart HC",
                "SA",
                "Random search",
            ],
            value=self.optimize_method,
            description="O. Type",
            disabled=False,
        )
        self.stepwise_path = False
        self.stepwise_checkbox = widgets.Checkbox(
            value=self.stepwise_path, description="Stepwise Path"
        )
        self.stepwise_checkbox.observe(self._animate_check_box_callback, names="value")
        self.next_step_button = widgets.Button(description="Next step")
        self.next_step_button.disabled = True
        self.initialize_bttn = widgets.Button(description="Initialize!")
        self.initialize_bttn.disabled = False
        self.get_neighbors_bttn = widgets.Button(description="Get Neighbors")
        self.get_neighbors_bttn.disabled = True
        self.select_successor_bttn = widgets.Button(description="Select Successor")
        self.select_successor_bttn.disabled = True
        self.optimize_bttn = widgets.Button(description="Auto optimize")
        self.optimize_bttn.disabled = True
        self.draw_best_bttn = widgets.Button(description="Draw best")
        self.draw_best_bttn.disabled = True
        config_box = widgets.HBox([self.tpye_dd, self.initialize_bttn])
        stepwise_box = widgets.HBox([self.stepwise_checkbox, self.next_step_button])

        self.initialize_bttn.on_click(self._init_optimization)
        self.next_step_button.on_click(self._next_path_segment)
        self.get_neighbors_bttn.on_click(self._get_neighborhood)
        self.select_successor_bttn.on_click(self._select_successor)
        self.optimize_bttn.on_click(self._optimize)
        self.draw_best_bttn.on_click(self._draw_best)
        self.tpye_dd.observe(self._optimize_method_dd_callback, names="value")

        items = [
            config_box,
            stepwise_box,
            self.get_neighbors_bttn,
            self.select_successor_bttn,
            self.optimize_bttn,
        ]
        opti_box = widgets.VBox(items)
        ui_box = widgets.HBox([opti_box, self.draw_best_bttn])

        display(ui_box, self.field_vis.simulation_canvas)

        self.field_vis.draw_optimization_frame(self.optimize_sim)

    def _init_optimization(self, _):
        self.optimize_sim.init_optimization(self.optimize_method)
        self.tpye_dd.disabled = True
        self.get_neighbors_bttn.disabled = False
        self.draw_best_bttn.disabled = False
        self.optimize_bttn.disabled = False
        self._update_frame()

    def _draw_best(self, _):
        if self.showing_best:
            self.field_vis.draw_optimization_frame(self.optimize_sim, draw_best=False)
            self.draw_best_bttn.description = "Draw best"
            self.showing_best = False
        else:
            self.field_vis.draw_optimization_frame(self.optimize_sim, draw_best=True)
            self.draw_best_bttn.description = "Draw current"
            self.showing_best = True

    def _optimize(self, _):
        while self.iteration < self.max_iterations:
            self.optimize_sim.create_neighborhood()
            terminate = self.optimize_sim.set_successor()
            if terminate:
                self._update_frame()
                break
            self.iteration += 1
        self.get_neighbors_bttn.disabled = True
        self.select_successor_bttn.disabled = True
        self.optimize_bttn.disabled = True
        if self.iteration == self.max_iterations:
            self._update_frame()

    def _next_path_segment(self, _):
        self.field_vis.draw_optimization_frame(
            simulation=self.optimize_sim, draw_idx=self.current_idx
        )
        self.current_idx += 1
        if self.current_idx >= len(self.optimize_sim.path):
            self.next_step_button.disabled = True

    def _select_successor(self, _):
        terminate = self.optimize_sim.set_successor()
        if terminate:
            self.get_neighbors_bttn.disabled = True
        else:
            self.get_neighbors_bttn.disabled = False
        self.select_successor_bttn.disabled = True
        self.optimize_bttn.disabled = True
        self.iteration += 1
        self._update_frame()

    def _update_frame(self):
        self.showing_best = False
        self.draw_best_bttn.description = "Draw best"
        self.current_idx = 0
        if self.stepwise_path:
            self.next_step_button.disabled = False

            self.field_vis.draw_optimization_frame(
                simulation=self.optimize_sim, draw_idx=self.current_idx
            )
            self.current_idx += 1
        else:
            self.field_vis.draw_optimization_frame(simulation=self.optimize_sim)

    def _get_neighborhood(self, _):
        self.optimize_sim.create_neighborhood()
        self.get_neighbors_bttn.disabled = True
        self.select_successor_bttn.disabled = False
        self.optimize_bttn.disabled = True

    def _animate_check_box_callback(self, msg):
        self.stepwise_path = msg["new"]
        self.current_idx = 0
        self.next_step_button.disabled = not self.stepwise_path
        if self.stepwise_path:
            self.field_vis.draw_optimization_frame(
                simulation=self.optimize_sim, draw_idx=0
            )
        else:
            self.field_vis.draw_optimization_frame(simulation=self.optimize_sim)

    def _optimize_method_dd_callback(self, msg):
        self.optimize_method = msg["new"]
