import networkx as nx
import numpy as np
import random
import itertools
import pickle

from ipycanvas import Canvas
from os.path import join
from matplotlib import pyplot as plt
import ipywidgets as widgets
from IPython.display import display
from PIL import Image as PILImage
from ipywidgets import Image


class Simulation:
    def __init__(
        self,
        graph_load_dir: str,
        t_schedule: callable,
        random_search: callable,
        stochastic_hc: callable,
        dummy: bool = False,
    ):
        self.path = []
        self.state = []
        self.last_state = []
        self.cost_history = []
        self.best_state = []
        self.best_state_cost = 0
        self.best_path = []
        self._load_graph(load_dir=graph_load_dir)
        if dummy:
            return
        self.neighbor_fig = plt.figure()
        self.neighbor_ax = self.neighbor_fig.add_subplot()
        self.t_schedule = t_schedule
        self.random_search = random_search
        self.stochastic_hc = stochastic_hc

    def _load_graph(self, load_dir: str):
        y_scale = 0.5
        with open(join(load_dir, "node_link_data435o.bin"), "rb") as f:
            node_link_data = pickle.load(f)
        self.sim_graph: nx.Graph = nx.node_link_graph(node_link_data)

        max_x = max(self.sim_graph.nodes()[n]["pos"][0] for n in self.sim_graph.nodes())
        self.sim_graph.add_node("Hof", pos=np.array([max_x, 0.0]))
        for node in list(self.sim_graph.nodes()):
            if node == "Hof":
                continue
            # self.sim_graph.add_edge("Hof", node)
            self.sim_graph.add_edge(node, "Hof")
            self.sim_graph.nodes()[node]["pos"][1] = (
                self.sim_graph.nodes()[node]["pos"][1] * y_scale
            )
        self.prepare_edge_costs()

    def prepare_edge_costs(self):
        for edge in self.sim_graph.edges():
            a = self.sim_graph.nodes()[edge[0]]["pos"] * 100
            b = self.sim_graph.nodes()[edge[1]]["pos"] * 100
            dist = int(np.linalg.norm(a - b).round(decimals=1))
            self.sim_graph.edges()[edge]["dist"] = dist

    def create_random_state(self) -> list[str, str]:
        state = []
        for node in list(self.sim_graph.nodes()):
            if node == "Hof":
                continue
            state.append(node)
        random.shuffle(state)
        state.insert(0, "Hof")
        state.insert(5, "Hof")
        return state

    def evaluate_state(self, state: list = None) -> float:
        if state is None:
            state = self.state
        self.path = []
        dist_sum = 0
        for idx, node in enumerate(state):
            if idx == (len(state) - 1):
                dist = self.sim_graph.edges()[state[idx], state[0]]["dist"]
                dist_sum = dist_sum + dist
            else:
                dist = self.sim_graph.edges()[state[idx], state[idx + 1]]["dist"]
                dist_sum = dist_sum + dist
        self.path.extend(state)
        return dist_sum

    def init_optimization(self, optimize_method: str, keep_history: bool = False):
        self.last_state = []
        self.optimize_method = optimize_method
        self.state = self.create_random_state()
        cost = self.evaluate_state()
        if not keep_history:
            self.best_path = self.path
            self.best_state_cost = cost
            self.best_state = self.state
            self.cost_history = []
            self.cost_history.append(round(cost, 2))

    def create_neighborhood(self):
        self.last_state = self.state
        combinations = list(itertools.combinations(range(0, len(self.state) - 2), 2))
        unique_combinations = list(set(combinations))
        random.shuffle(unique_combinations)
        self.draw_neighborhood(neighborhood=unique_combinations)
        self.neighborhood = unique_combinations

    def draw_neighborhood(self, neighborhood: list):
        plt.cla()
        costs = []
        neighbor_idcs = []
        for neighbor in neighborhood:
            neighbor_state, idcs = self.apply_neighbor_to_state(
                neighbor, return_idcs=True
            )
            cost = self.evaluate_state(state=neighbor_state)
            costs.append(cost)
            neighbor_idcs.append(idcs)
        costs = np.array(costs)

        self.neighbor_ax.scatter(0, self.evaluate_state(self.state), c="tab:red")
        for idx, cost in zip(range(len(neighborhood)), reversed(costs)):
            self.neighbor_ax.bar(idx + 1, cost, color="tab:blue")
        # min line
        self.neighbor_ax.plot(
            [0, len(neighborhood) + 1],
            [costs.min(), costs.min()],
            "b:",
        )
        # current line
        current_state_cost = self.evaluate_state(self.state)
        self.neighbor_ax.plot(
            [0, len(neighborhood) + 1],
            [current_state_cost, current_state_cost],
            "r:",
        )
        # plot cost history
        cost_plot_idcs = [
            i + len(neighborhood) + 1 for i in range(len(self.cost_history))
        ]
        self.neighbor_ax.plot(cost_plot_idcs, self.cost_history)

        # x-axis labels
        neighborhood_labels = [
            (self.state[idx1], self.state[idx2])
            for idx1, idx2 in reversed(neighbor_idcs)
        ]

        self.neighbor_ax.set_xticks(
            list(range(len(neighborhood) + 1 + len(self.cost_history))),
            ["current", *neighborhood_labels, *list(range(len(self.cost_history)))],
            rotation=90,
        )

        self.neighbor_fig.suptitle("Neighborhood & cost history")
        self.neighbor_ax.set_ylabel("Cost")
        self.neighbor_fig.subplots_adjust(bottom=0.35)

    def apply_neighbor_to_state(
        self, neighbor: list[int, int], return_idcs: bool = False
    ) -> list:
        state = self.state.copy()
        node1, node2 = neighbor
        idx1 = state.index(node1)
        idx2 = state.index(node2)
        old_val_idx1 = state[idx1]
        state[idx1] = state[idx2]
        state[idx2] = old_val_idx1
        if return_idcs:
            return state, (idx1, idx2)
        else:
            return state

    def hc_successor(self) -> [list, bool]:
        terminate = True
        cost = self.evaluate_state()
        selected_neighbor_state = self.state
        for neighbor in self.neighborhood:
            neighbor_state = self.apply_neighbor_to_state(neighbor=neighbor)
            neighbor_cost = self.evaluate_state(state=neighbor_state)
            if neighbor_cost < cost:
                cost = neighbor_cost
                selected_neighbor_state = neighbor_state
                terminate = False

        return selected_neighbor_state, terminate

    def first_choice_hc_successor(self) -> [list, bool]:
        cost = self.cost_history[-1]
        local_neighborhood = self.neighborhood.copy()
        while cost >= self.cost_history[-1]:
            neighbor = local_neighborhood.pop()
            neighbor_state = self.apply_neighbor_to_state(neighbor=neighbor)
            cost = self.evaluate_state(state=neighbor_state)
            if not local_neighborhood:
                return self.state, True

        return neighbor_state, False

    def stochastic_hc_successor(self) -> [list, bool]:
        neighbor_states = []
        neighbor_costs = []
        for neighbor in self.neighborhood:
            neighbor_state = self.apply_neighbor_to_state(neighbor=neighbor)
            neighbor_cost = self.evaluate_state(state=neighbor_state)
            neighbor_costs.append(neighbor_cost)
            neighbor_states.append(neighbor_state)
        return self.stochastic_hc(
            current_state=self.state,
            neighbor_states=neighbor_states,
            neighbor_costs=neighbor_costs,
        )

    def random_restart_hc_successor(self) -> [list, bool]:
        hc_successor, terminate = self.hc_successor()
        if terminate:
            # perform restart
            self.init_optimization(
                optimize_method=self.optimize_method, keep_history=True
            )
            hc_successor = self.state
        return hc_successor, False

    def sa_successor(self) -> [list, bool]:
        # get t
        t = len(self.cost_history)
        # calculate temperature with function of user
        temperature = self.t_schedule(t) * 100
        if temperature <= 0:
            return self.state, True
        # choose random successor
        local_neighborhood = self.neighborhood.copy()
        neighbor_candidate = local_neighborhood.pop()
        neighbor_candidate_state = self.apply_neighbor_to_state(
            neighbor=neighbor_candidate
        )
        # get cost of the successor
        cost = self.evaluate_state(state=neighbor_candidate_state)

        delta_e = cost - self.cost_history[-1]
        if delta_e < 0:
            return neighbor_candidate_state, False
        else:
            propability = np.exp(-delta_e / temperature)
            print(
                f"Keine Verbesserung mit delta_e {delta_e:.2f}, temperature {temperature / 100:.2f} und Annahmewahrscheinlichkeit {propability * 100:.2f}%"
            )
            if random.random() <= propability:
                return neighbor_candidate_state, False
            else:
                return self.state, False

    def random_search_successor(self) -> [list, bool]:
        neighbor_states = []
        neighbor_costs = []
        for neighbor in self.neighborhood:
            neighbor_state = self.apply_neighbor_to_state(neighbor=neighbor)
            neighbor_cost = self.evaluate_state(state=neighbor_state)
            neighbor_costs.append(neighbor_cost)
            neighbor_states.append(neighbor_state)
        return self.random_search(
            current_state=self.state,
            neighbor_states=neighbor_states,
            neighbor_costs=neighbor_costs,
        )

    def set_successor(self):
        # method
        if self.optimize_method == "Hill climbing":
            chosen_state, terminate = self.hc_successor()
        elif self.optimize_method == "First-choice HC":
            chosen_state, terminate = self.first_choice_hc_successor()
        elif self.optimize_method == "Stochastic HC":
            chosen_state, terminate = self.stochastic_hc_successor()
        elif self.optimize_method == "Random-restart HC":
            chosen_state, terminate = self.random_restart_hc_successor()
        elif self.optimize_method == "SA":
            chosen_state, terminate = self.sa_successor()
        elif self.optimize_method == "Random search":
            chosen_state, terminate = self.random_search_successor()
        else:
            return

        self.state = chosen_state

        cost = self.evaluate_state()
        if cost < self.best_state_cost:
            self.best_state_cost = cost
            self.best_state = self.state
            self.best_path = self.path
        if not terminate:
            self.cost_history.append(round(cost, 2))
        return terminate


class Visualizer:
    def __init__(self, simulation: Simulation, canvas_size: int = 400):
        self.sim_canvas_size = canvas_size
        self.sim_canvas_width = int(self.sim_canvas_size * 16 / 9)
        self.simulation_canvas = Canvas(
            height=self.sim_canvas_size, width=self.sim_canvas_width
        )
        self._calculate_scaling_params_for_simulation(simulation=simulation)
        radius = 20
        self.straw_sprite = PILImage.open("sprites/straw.png")
        self.straw_sprite.thumbnail((radius, radius), PILImage.Resampling.LANCZOS)
        self.straw_sprite.save("sprites/straw_8.png", quality=10)
        self.barn_sprite = PILImage.open("sprites/barn.png")
        self.barn_sprite.thumbnail(
            (radius * 4, radius * 4), PILImage.Resampling.LANCZOS
        )
        self.barn_sprite.save("sprites/barn_8.png", quality=10)
        self.straw_sprite = Image.from_file("sprites/straw_8.png")
        self.barn_sprite = Image.from_file("sprites/barn_8.png")

    def _calculate_scaling_params_for_simulation(self, simulation: Simulation):
        xs = np.array(
            [
                simulation.sim_graph.nodes()[n]["pos"]
                for n in simulation.sim_graph.nodes()
            ]
        )
        ys = np.array(
            [
                simulation.sim_graph.nodes()[n]["pos"][1]
                for n in simulation.sim_graph.nodes()
            ]
        )

        min_x = xs.min() - 0.2  # provisorische Lösung für Node am Rand wird halbiert
        max_x = xs.max() + 0.2
        min_y = ys.min() - 0.2
        max_y = ys.max() + 0.2

        delta_x = max_x - min_x
        delta_y = max_y - min_y

        if delta_x > delta_y:
            high_delta = delta_x
        else:
            high_delta = delta_y

        self.sim_canvas_factor = self.sim_canvas_size / high_delta
        self.sim_canvas_x_offset = (
            abs(min_x) + (self.sim_canvas_size / self.sim_canvas_factor - delta_x) / 2
        )
        self.sim_canvas_y_offset = (
            abs(min_y) + (self.sim_canvas_size / self.sim_canvas_factor - delta_y) / 2
        )

    def _rescale_to_simulation_frame_coords(
        self, coord: np.array
    ) -> tuple[np.array, np.array]:
        return (coord[0] + self.sim_canvas_x_offset) * self.sim_canvas_factor, (
            coord[1] + self.sim_canvas_y_offset
        ) * self.sim_canvas_factor

    def draw_nodes(self, simulation: Simulation):
        for node in list(simulation.sim_graph.nodes()):
            pos = simulation.sim_graph.nodes()[node]["pos"]
            scaled_pos_x, scaled_pos_y = self._rescale_to_simulation_frame_coords(
                np.array((pos[0], pos[1]))
            )
            if type(node) != int:
                self.simulation_canvas.draw_image(
                    self.barn_sprite, scaled_pos_x, scaled_pos_y
                )
            else:
                self.simulation_canvas.draw_image(
                    self.straw_sprite, scaled_pos_x, scaled_pos_y
                )

            self.simulation_canvas.fill_style = "#000000"
            self.simulation_canvas.fill_text(
                text=node,
                x=scaled_pos_x,
                y=scaled_pos_y,
            )

    @staticmethod
    def rgb2hex(r, g, b):
        return "#{:02x}{:02x}{:02x}".format(r, g, b)

    def draw_current_solution_path(
        self, simulation: Simulation, draw_best: bool = False
    ):
        if draw_best:
            path = simulation.best_path
        else:
            path = simulation.path
        self.simulation_canvas.line_width = 3

        for idx, node in enumerate(path):
            r = 90
            g = 120
            b = 90
            if idx > 5:
                r = 190

            if idx == 0:
                pos1 = simulation.sim_graph.nodes()[path[len(path) - 1]]["pos"]
                pos2 = simulation.sim_graph.nodes()[node]["pos"]
                r = 190
            else:
                pos1 = simulation.sim_graph.nodes()[path[idx - 1]]["pos"]
                pos2 = simulation.sim_graph.nodes()[node]["pos"]

            rescaled_node_1 = self._rescale_to_simulation_frame_coords(
                np.array((pos1[0], pos1[1]))
            )
            rescaled_node_2 = self._rescale_to_simulation_frame_coords(
                np.array((pos2[0], pos2[1]))
            )
            self.simulation_canvas.stroke_style = self.rgb2hex(r, g, b)
            self.simulation_canvas.begin_path()
            self.simulation_canvas.move_to(rescaled_node_1[0], rescaled_node_1[1])
            self.simulation_canvas.line_to(rescaled_node_2[0], rescaled_node_2[1])
            self.simulation_canvas.stroke()

    def draw_path_segment(
        self, simulation: Simulation, idx: int, draw_best: bool = False
    ):
        if draw_best:
            path = simulation.best_path
        else:
            path = simulation.path
        self.simulation_canvas.line_width = 3
        if idx == 0:
            return
        node = path[idx]
        r = int(idx / len(path) * 200)
        g = 200 - int(idx / len(path) * 200)
        self.simulation_canvas.stroke_style = self.rgb2hex(r, g, 140)
        edge = (path[idx - 1], node)
        dubin_x = simulation.sim_graph.edges()[edge]["path_x"]
        dubin_y = simulation.sim_graph.edges()[edge]["path_y"]
        rescaled_path = self._rescale_to_simulation_frame_coords(
            np.array([dubin_x, dubin_y])
        )
        lines = [np.array((x, y)) for x, y in zip(rescaled_path[0], rescaled_path[1])]
        self.simulation_canvas.begin_path()
        self.simulation_canvas.move_to(*lines[0])
        for p in lines:
            self.simulation_canvas.line_to(*p)

        self.simulation_canvas.stroke()

    def draw_available_path(self, simulation: Simulation, current_node: str):
        self.simulation_canvas.line_width = 3
        self.simulation_canvas.stroke_style = "green"
        for edge in list(simulation.sim_graph.edges(current_node)):
            dubin_x = simulation.sim_graph.edges()[edge]["path_x"]
            dubin_y = simulation.sim_graph.edges()[edge]["path_y"]
            rescaled_path = self._rescale_to_simulation_frame_coords(
                np.array([dubin_x, dubin_y])
            )
            lines = [
                np.array((x, y)) for x, y in zip(rescaled_path[0], rescaled_path[1])
            ]
            self.simulation_canvas.begin_path()
            self.simulation_canvas.move_to(*lines[0])
            for p in lines:
                self.simulation_canvas.line_to(*p)
            self.simulation_canvas.stroke()

    def draw_states(self, simulation: Simulation):
        # clear info side
        self.simulation_canvas.clear_rect(
            self.sim_canvas_size + 10, 0, self.sim_canvas_size, self.sim_canvas_size
        )
        self.simulation_canvas.fill_style = "#000000"
        self.simulation_canvas.font = "15px Arial"

        self.simulation_canvas.fill_text(
            text="last state",
            x=self.sim_canvas_size + 10,
            y=20,
        )
        self.simulation_canvas.fill_text(
            text="current state",
            x=self.sim_canvas_size + 100,
            y=20,
        )
        self.simulation_canvas.fill_text(
            text="cost history",
            x=self.sim_canvas_size + 200,
            y=20,
        )
        self.simulation_canvas.fill_text(
            text="best",
            x=self.sim_canvas_size + 300,
            y=20,
        )
        # find indices of changed rows
        changed_state_idx = []
        if simulation.last_state:
            changed_state_idx = [
                idx
                for idx, item in enumerate(simulation.state)
                if item != simulation.last_state[idx]
            ]

        for idx, row in enumerate(simulation.last_state):
            self.simulation_canvas.fill_text(
                text=row,
                x=self.sim_canvas_size + 10,
                y=50 + idx * 16,
            )

        for idx, row in enumerate(simulation.state):
            # highlight changed rows
            if idx in changed_state_idx:
                self.simulation_canvas.fill_style = "#9F0000"
            self.simulation_canvas.fill_text(
                text=row,
                x=self.sim_canvas_size + 100,
                y=50 + idx * 16,
            )
            self.simulation_canvas.fill_style = "#000000"

        # draw cost history
        if len(simulation.cost_history) > 20:
            cost_hist = simulation.cost_history[-20:]
        else:
            cost_hist = simulation.cost_history
        for idx, cost in enumerate(cost_hist):
            self.simulation_canvas.fill_text(
                text=cost,
                x=self.sim_canvas_size + 200,
                y=50 + idx * 16,
            )
        # draw best state
        for idx, row in enumerate(simulation.best_state):
            self.simulation_canvas.fill_text(
                text=row,
                x=self.sim_canvas_size + 300,
                y=50 + idx * 16,
            )
        # best cost
        self.simulation_canvas.fill_text(
            text=f"cost={simulation.best_state_cost:.2f}",
            x=self.sim_canvas_size + 300,
            y=50 + len(simulation.best_state) * 16,
        )

    def draw_optimization_frame(
        self, simulation: Simulation, draw_idx: int = None, draw_best: bool = False
    ):
        if draw_idx is None or draw_idx == 0:
            self.draw_background()
            self.draw_nodes(simulation=simulation)
            self.draw_states(simulation=simulation)
        if draw_idx is None:
            self.draw_current_solution_path(simulation=simulation, draw_best=draw_best)
        else:
            self.draw_path_segment(
                simulation=simulation, idx=draw_idx, draw_best=draw_best
            )

    def draw_background(self):
        self.simulation_canvas.clear_rect(
            0, 0, self.sim_canvas_width, self.sim_canvas_size
        )


class IPyUI:
    def __init__(
        self, field_vis: Visualizer, optimize_sim: Simulation, max_iterations: int
    ):
        self.field_vis = field_vis
        self.optimize_sim = optimize_sim
        self.current_idx = 0
        self.showing_best = False
        self.max_iterations = max_iterations
        self.iteration = 0

    def optimization_ui(self):
        # Optimization settings
        self.optimize_method = "Hill climbing"
        self.tpye_dd = widgets.Dropdown(
            options=[
                "Hill climbing",
                "First-choice HC",
                "Stochastic HC",
                "Random-restart HC",
                "SA",
                "Random search",
            ],
            value=self.optimize_method,
            description="O. Type",
            disabled=False,
        )
        self.stepwise_path = False
        self.stepwise_checkbox = widgets.Checkbox(
            value=self.stepwise_path, description="Stepwise Path"
        )
        self.stepwise_checkbox.disabled = True
        self.stepwise_checkbox.observe(self._animate_check_box_callback, names="value")
        self.next_step_button = widgets.Button(description="Next step")
        self.next_step_button.disabled = True
        self.initialize_bttn = widgets.Button(description="Initialize!")
        self.initialize_bttn.disabled = False
        self.get_neighbors_bttn = widgets.Button(description="Get Neighbors")
        self.get_neighbors_bttn.disabled = True
        self.select_successor_bttn = widgets.Button(description="Select Successor")
        self.select_successor_bttn.disabled = True
        self.optimize_bttn = widgets.Button(description="Auto optimize")
        self.optimize_bttn.disabled = True
        self.draw_best_bttn = widgets.Button(description="Draw best")
        self.draw_best_bttn.disabled = True
        config_box = widgets.HBox([self.tpye_dd, self.initialize_bttn])
        stepwise_box = widgets.HBox([self.stepwise_checkbox, self.next_step_button])

        self.initialize_bttn.on_click(self._init_optimization)
        self.next_step_button.on_click(self._next_path_segment)
        self.get_neighbors_bttn.on_click(self._get_neighborhood)
        self.select_successor_bttn.on_click(self._select_successor)
        self.optimize_bttn.on_click(self._optimize)
        self.draw_best_bttn.on_click(self._draw_best)
        self.tpye_dd.observe(self._optimize_method_dd_callback, names="value")

        items = [
            config_box,
            stepwise_box,
            self.get_neighbors_bttn,
            self.select_successor_bttn,
            self.optimize_bttn,
        ]
        opti_box = widgets.VBox(items)
        ui_box = widgets.HBox([opti_box, self.draw_best_bttn])

        display(ui_box, self.field_vis.simulation_canvas)

        self.field_vis.draw_optimization_frame(self.optimize_sim)

    def _init_optimization(self, _):
        self.optimize_sim.init_optimization(self.optimize_method)
        self.tpye_dd.disabled = True
        self.get_neighbors_bttn.disabled = False
        self.draw_best_bttn.disabled = False
        self.optimize_bttn.disabled = False
        self._update_frame()

    def _draw_best(self, _):
        if self.showing_best:
            self.field_vis.draw_optimization_frame(self.optimize_sim, draw_best=False)
            self.draw_best_bttn.description = "Draw best"
            self.showing_best = False
        else:
            self.field_vis.draw_optimization_frame(self.optimize_sim, draw_best=True)
            self.draw_best_bttn.description = "Draw current"
            self.showing_best = True

    def _optimize(self, _):
        while self.iteration < self.max_iterations:
            self.optimize_sim.create_neighborhood()
            terminate = self.optimize_sim.set_successor()
            if terminate:
                self._update_frame()
                break
            self.iteration += 1
        self.get_neighbors_bttn.disabled = True
        self.select_successor_bttn.disabled = True
        self.optimize_bttn.disabled = True
        if self.iteration == self.max_iterations:
            self._update_frame()

    def _next_path_segment(self, _):
        self.field_vis.draw_optimization_frame(
            simulation=self.optimize_sim, draw_idx=self.current_idx
        )
        self.current_idx += 1
        if self.current_idx >= len(self.optimize_sim.path):
            self.next_step_button.disabled = True

    def _select_successor(self, _):
        terminate = self.optimize_sim.set_successor()
        if terminate:
            self.get_neighbors_bttn.disabled = True
        else:
            self.get_neighbors_bttn.disabled = False
        self.select_successor_bttn.disabled = True
        self.optimize_bttn.disabled = True
        self.iteration += 1
        self._update_frame()

    def _update_frame(self):
        self.showing_best = False
        self.draw_best_bttn.description = "Draw best"
        self.current_idx = 0
        if self.stepwise_path:
            self.next_step_button.disabled = False

            self.field_vis.draw_optimization_frame(
                simulation=self.optimize_sim, draw_idx=self.current_idx
            )
            self.current_idx += 1
        else:
            self.field_vis.draw_optimization_frame(simulation=self.optimize_sim)

    def _get_neighborhood(self, _):
        self.optimize_sim.create_neighborhood()
        self.get_neighbors_bttn.disabled = True
        self.select_successor_bttn.disabled = False
        self.optimize_bttn.disabled = True

    def _animate_check_box_callback(self, msg):
        self.stepwise_path = msg["new"]
        self.current_idx = 0
        self.next_step_button.disabled = not self.stepwise_path
        if self.stepwise_path:
            self.field_vis.draw_optimization_frame(
                simulation=self.optimize_sim, draw_idx=0
            )
        else:
            self.field_vis.draw_optimization_frame(simulation=self.optimize_sim)

    def _optimize_method_dd_callback(self, msg):
        self.optimize_method = msg["new"]
