import json
import networkx as nx
import numpy as np
import pickle

from ipycanvas import Canvas
from os.path import join
from natsort import natsorted
import ipywidgets as widgets
from IPython.display import display

CURRENT_COLOR = "#F03000"
GOAL_COLOR = "#800080"
EXPLORED_COLOR = "#00a000"
IN_FRONTIER_COLOR = "#528afe"
NORMAL_COLOR = "#a0a0a0"


class Simulation:
    def __init__(self, load_dir: str):
        self.explored_set = set()
        self.frontier = []
        self.frontier_set = set(self.frontier)
        self.frontier_history = {
            "item": [],
            "avail": [],
            "cost": [],
        }
        self.frontier_history_unsorted = {
            "item": [],
            "step": [],
            "cost": [],
        }
        self.expansion_step = 0
        self.path = []
        self.goal_node = 0

        self.load_graph(load_dir=load_dir)

    def load_graph(self, load_dir: str):
        with open(join(load_dir, "node_link_data.bin"), "rb") as f:
            node_link_data = pickle.load(f)
        self.sim_graph: nx.Graph = nx.node_link_graph(node_link_data)
        self.prepare_edge_costs()

    def init_search(self, start_node: str, search_type: str, search_method: str):
        self.explored_set = set()
        self.frontier_set = set()
        self.frontier = []
        self.frontier_history = {
            "item": [],
            "avail": [],
            "cost": [],
            "path_cost": [],
            "heuristic": [],
        }
        self.frontier_history_unsorted = {
            "item": [],
            "step": [],
            "cost": [],
            "path_cost": [],
            "heuristic": [],
        }
        self.path = []

        self.chosen = start_node
        self.search_type = search_type
        self.search_method = search_method

    def set_goal_node(self, goal_node: int):
        self.goal_node = goal_node

    def prepare_edge_costs(self):
        for edge in self.sim_graph.edges():
            a = self.sim_graph.nodes()[edge[0]]["pos"] * 100
            b = self.sim_graph.nodes()[edge[1]]["pos"] * 100
            dist = int(np.linalg.norm(a - b).round(decimals=1))
            self.sim_graph.edges()[edge]["dist"] = dist

    def choose(self) -> tuple[str, str]:
        if not self.frontier:
            return None
        if self.search_method == "DFS":
            return self.choose_queue()
        else:
            return self.choose_stack()

    def choose_stack(self) -> tuple[str, str]:
        self.chosen = self.frontier.pop(0)[1]
        found_in_history = self.mark_as_not_avail_in_history(self.chosen)
        if found_in_history:
            return self.chosen
        else:
            return None

    def choose_queue(self) -> tuple[str, str]:
        self.chosen = self.frontier.pop()[1]
        found_in_history = self.mark_as_not_avail_in_history(self.chosen)
        if found_in_history:
            return self.chosen
        else:
            return None

    def mark_as_not_avail_in_history(self, chosen_node) -> bool:
        for idx, item in reversed(list(enumerate(self.frontier_history["item"]))):
            if item[1] == chosen_node and self.frontier_history["avail"][idx]:
                self.frontier_history["avail"][idx] = False
                return True
        return False

    def mark_expansion_step(self, chosen_node):
        for idx, item in reversed(
            list(enumerate(self.frontier_history_unsorted["item"]))
        ):
            if item[1] == chosen_node:
                self.frontier_history_unsorted["step"][idx] = self.expansion_step

    def generate_child_nodes(self, node: str) -> list[str]:
        return [n for n in self.sim_graph.neighbors(node)]

    def expand(self) -> tuple[bool, str]:
        ### this has grown very badly ### Please rewrite
        self.current_node = self.chosen
        self.mark_expansion_step(self.current_node)
        self.expansion_step += 1
        self.chosen = None
        goal_reached = False
        goal_node = ""
        if self.search_type == "graph":
            self.explored_set.add(self.current_node)
        children = sorted(self.generate_child_nodes(self.current_node))
        for child in children:
            # check if child already in explored or frontier

            if child in self.explored_set:
                continue

            # calculate the path cost g(n) and heuristic values h(n)
            cost_to_node = self.get_cost_of_current_path()
            cost_from_node_to_child = self.sim_graph.edges()[
                (self.current_node, child)
            ]["dist"]
            path_cost = cost_to_node + cost_from_node_to_child
            heuristic = self.get_heuristic(node=child)

            # check if cheaper version is available if already in frontier
            if child in self.frontier_set:
                # when using BFS or DFS continue without checking
                if self.search_method == "DFS" or self.search_method == "BFS":
                    continue
                hist_idx = next(
                    i
                    for i, (*_, v) in enumerate(self.frontier_history["item"])
                    if v == child
                )
                hist_unsorted_idx = next(
                    i
                    for i, (*_, v) in enumerate(self.frontier_history_unsorted["item"])
                    if v == child
                )

                if self.search_method == "UCS":
                    cost = path_cost
                    if self.frontier_history["cost"][hist_idx] <= path_cost:
                        continue
                if self.search_method == "GreedyBestFirst":
                    cost = heuristic
                    if self.frontier_history["cost"][hist_idx] <= heuristic:
                        continue
                if self.search_method == "A*":
                    cost = heuristic + path_cost
                    if self.frontier_history["cost"][hist_idx] <= path_cost + heuristic:
                        continue

                # update entry in frontier_hist
                self.frontier_history["item"][hist_idx] = (self.current_node, child)
                self.frontier_history_unsorted["item"][hist_unsorted_idx] = (
                    self.current_node,
                    child,
                )
                self.frontier_history["avail"][hist_idx] = True
                self.frontier_history_unsorted["step"][hist_unsorted_idx] = ""

                self.frontier_history["path_cost"][hist_idx] = path_cost
                self.frontier_history_unsorted["path_cost"][
                    hist_unsorted_idx
                ] = path_cost
                self.frontier_history["heuristic"][hist_idx] = heuristic
                self.frontier_history_unsorted["heuristic"][
                    hist_unsorted_idx
                ] = heuristic

                self.frontier_history["cost"][hist_idx] = cost
                self.frontier_history_unsorted["cost"][hist_unsorted_idx] = cost
            else:
                # create entry in frontier
                self.frontier.append((self.current_node, child))
                self.frontier_history["item"].append((self.current_node, child))
                self.frontier_history_unsorted["item"].append(
                    (self.current_node, child)
                )
                self.frontier_history["avail"].append(True)
                self.frontier_history_unsorted["step"].append("")

                self.frontier_history["path_cost"].append(path_cost)
                self.frontier_history_unsorted["path_cost"].append(path_cost)
                self.frontier_history["heuristic"].append(heuristic)
                self.frontier_history_unsorted["heuristic"].append(heuristic)

                # calculate f(n)
                if self.search_method == "UCS":
                    # f(n) = g(n)
                    self.frontier_history["cost"].append(path_cost)
                    self.frontier_history_unsorted["cost"].append(path_cost)
                elif self.search_method == "GreedyBestFirst":
                    # f(n) = h(n)
                    self.frontier_history["cost"].append(heuristic)
                    self.frontier_history_unsorted["cost"].append(heuristic)
                elif self.search_method == "A*":
                    # f(n) = g(n) + h(n)
                    self.frontier_history["cost"].append(path_cost + heuristic)
                    self.frontier_history_unsorted["cost"].append(path_cost + heuristic)
                else:
                    self.frontier_history["cost"].append(path_cost)
                    self.frontier_history_unsorted["cost"].append(path_cost)

            if self.search_method == "DFS" or self.search_method == "BFS":
                # # test every child added to frontier for goal
                if not goal_reached:
                    goal_reached = self.goal_test(child)
                    if goal_reached:
                        goal_node = child

        # sort frontier if neccessary
        self.frontier_set = set([item[1] for item in self.frontier_history["item"]])
        if (
            self.search_method == "UCS"
            or self.search_method == "A*"
            or self.search_method == "GreedyBestFirst"
        ):
            self.sort_frontier_and_hist()

        return goal_reached, goal_node

    def goal_test(self, node: str) -> bool:
        if self.goal_node == node:
            return True
        else:
            return False

    def sort_frontier_and_hist(self):
        items = np.array(self.frontier_history["item"])
        avails = np.array(self.frontier_history["avail"])
        costs = np.array(self.frontier_history["cost"])
        path_costs = np.array(self.frontier_history["path_cost"])
        heuristics = np.array(self.frontier_history["heuristic"])
        avails_mask = np.where(avails)
        not_avails_mask = np.where(avails == False)
        avail_frontier_hist = (
            items[avails_mask].tolist(),
            avails[avails_mask].tolist(),
            costs[avails_mask].tolist(),
            path_costs[avails_mask].tolist(),
            heuristics[avails_mask].tolist(),
        )
        not_avail_frontier_hist = (
            items[not_avails_mask].tolist(),
            avails[not_avails_mask].tolist(),
            costs[not_avails_mask].tolist(),
            path_costs[not_avails_mask].tolist(),
            heuristics[not_avails_mask].tolist(),
        )
        sorted_avail_frontier_history = [
            (item, avail, cost, path_cost, heuristic)
            for item, avail, cost, path_cost, heuristic in sorted(
                zip(*avail_frontier_hist), key=lambda pair: pair[2]
            )
        ]
        self.frontier_history["item"] = not_avail_frontier_hist[0]
        self.frontier_history["avail"] = not_avail_frontier_hist[1]
        self.frontier_history["cost"] = not_avail_frontier_hist[2]
        self.frontier_history["path_cost"] = not_avail_frontier_hist[3]
        self.frontier_history["heuristic"] = not_avail_frontier_hist[4]

        self.frontier = []
        for hist in sorted_avail_frontier_history:
            self.frontier.append(hist[0])
            self.frontier_history["item"].append(hist[0])
            self.frontier_history["avail"].append(hist[1])
            self.frontier_history["cost"].append(hist[2])
            self.frontier_history["path_cost"].append(hist[3])
            self.frontier_history["heuristic"].append(hist[4])

    def get_cost_of_current_path(self) -> float:
        cost = 0
        for edge in self.path:
            cost += self.sim_graph.edges()[edge]["dist"]
        return cost

    def get_heuristic(self, node: str):
        a = np.array(self.sim_graph.nodes()[node]["pos"][:2])
        b = np.array(self.sim_graph.nodes()[self.goal_node]["pos"][:2])
        dist = np.linalg.norm(a - b).item() * 100
        return round(dist, ndigits=0)

    def create_current_path(self, last_node: str):
        parent_available = True
        self.path = []
        while parent_available:
            parent_available = False
            for idx, item in enumerate(self.frontier_history["item"]):
                if item[1] == last_node:
                    last_node = self.frontier_history["item"][idx][0]
                    parent_available = True
                    self.path.append((item[1], item[0]))


class FieldSimulation(Simulation):
    def __init__(self, load_dir):
        super().__init__(load_dir)
        self.goal_node = "logistic0"

    def load_graph(self, load_dir: str):
        with open(join(load_dir, "node_link_data.json"), "r") as f:
            node_link_data = json.load(f)
        self.sim_graph: nx.Graph = nx.node_link_graph(node_link_data)

        with open(join(load_dir, "field_info.json"), "r") as f:
            self.field_info_dict = json.load(f)

        with open(join(load_dir, "field.json"), "r") as f:
            field_dict = json.load(f)

        self._prepare_field_for_sim(field_dict)

    def _prepare_field_for_sim(self, field_dict: dict):
        # get border from field dict
        self.field_border = field_dict["field_border"]
        # save logistic point names
        n_logistic_points = 1
        # n_logistic_points = self.logistic_point_attributes["quantity"]
        self.logistic_point_names = []
        for i in range(n_logistic_points):
            self.logistic_point_names.append(f"logistic{i}")

        # simplify graph
        simple_graph = nx.Graph()
        for node in list(self.sim_graph.nodes()):
            for edge in self.sim_graph.edges(node):
                simple_edge = (
                    self.simplify_node_name(edge[0]),
                    self.simplify_node_name(edge[1]),
                )
                if simple_edge in list(simple_graph.edges()):
                    continue
                simple_graph.add_node(
                    simple_edge[0], pos=self.sim_graph.nodes()[node]["pos"]
                )
                path_x = self.sim_graph.edges()[edge]["path_x"]
                path_y = self.sim_graph.edges()[edge]["path_y"]
                dist = self.sim_graph.edges()[edge]["dist"]
                headland = self.sim_graph.edges()[edge]["headland"]
                simple_graph.add_edge(
                    *simple_edge,
                    dist=dist,
                    headland=headland,
                    path_x=path_x,
                    path_y=path_y,
                )
        self.sim_graph = simple_graph

        # Create sub_graph for navigation only on headlands
        all_edges_list = list(self.sim_graph.edges)
        headland_edges = [
            e for e in all_edges_list if self.sim_graph.edges()[e]["headland"] == True
        ]
        self.headland_subgraph = self.sim_graph.edge_subgraph(headland_edges)

    def goal_test(self, node: str) -> bool:
        if "logistic" in node:
            return True
        else:
            return False

    def generate_child_nodes(self, node: str) -> list[str]:
        return self.headland_subgraph.neighbors(node)

    @staticmethod
    def simplify_node_name(node: str) -> str:
        node = node.replace("anticlock_", "")
        node = node.replace("clock_", "")
        node = node.replace("_in", "")
        node = node.replace("_out", "")
        return node


class Visualizer:
    def __init__(self, simulation: Simulation, canvas_size: int = 400):
        self.sim_canvas_size = canvas_size
        self.simulation_canvas = Canvas(
            height=self.sim_canvas_size, width=int(self.sim_canvas_size * 16 / 9)
        )
        self._calculate_scaling_params_for_simulation(simulation=simulation)

    def _calculate_scaling_params_for_simulation(self, simulation: Simulation):
        xs = np.array(
            [
                simulation.sim_graph.nodes()[n]["pos"]
                for n in simulation.sim_graph.nodes()
            ]
        )
        ys = np.array(
            [
                simulation.sim_graph.nodes()[n]["pos"][1]
                for n in simulation.sim_graph.nodes()
            ]
        )

        min_x = xs.min()
        max_x = xs.max()
        min_y = ys.min()
        max_y = ys.max()

        delta_x = max_x - min_x
        delta_y = max_y - min_y

        if delta_x > delta_y:
            high_delta = delta_x
        else:
            high_delta = delta_y

        self.sim_canvas_factor = self.sim_canvas_size / high_delta
        self.sim_canvas_x_offset = (
            abs(min_x) + (self.sim_canvas_size / self.sim_canvas_factor - delta_x) / 2
        )
        self.sim_canvas_y_offset = (
            abs(min_y) + (self.sim_canvas_size / self.sim_canvas_factor - delta_y) / 2
        )

    def _rescale_to_simulation_frame_coords(
        self, coord: np.array
    ) -> tuple[np.array, np.array]:
        return (coord[0] + self.sim_canvas_x_offset) * self.sim_canvas_factor, (
            coord[1] + self.sim_canvas_y_offset
        ) * self.sim_canvas_factor

    def get_node_info(self, simulation: Simulation, node: str) -> dict:
        edges = list(simulation.sim_graph.edges(node))
        costs = [simulation.sim_graph.edges()[e]["dist"] for e in edges]
        info = {
            "node": node,
            "edges": edges,
            "costs": costs,
        }
        return info

    def write_node_info_table(self, simulation: Simulation, current_node: str):
        if current_node is None:
            return

        # set draw parameters
        self.simulation_canvas.fill_style = "#000000"
        font_size = 15
        self.simulation_canvas.font = f"{font_size}px Arial"
        self.simulation_canvas.stroke_style = "#000000"
        self.simulation_canvas.line_width = 1
        header_y_offset = 20

        # clear child nodes, distances and explored set
        self.simulation_canvas.clear_rect(
            self.sim_canvas_size + 10, 0, self.sim_canvas_size, self.sim_canvas_size
        )
        node_info = self.get_node_info(simulation=simulation, node=current_node)
        # write headers
        self.simulation_canvas.fill_text(
            text="child nodes",
            x=self.sim_canvas_size + 10,
            y=header_y_offset,
        )
        self.simulation_canvas.fill_text(
            text="distance",
            x=self.sim_canvas_size + 100,
            y=header_y_offset,
        )
        self.simulation_canvas.fill_text(
            text="explored set",
            x=self.sim_canvas_size + 180,
            y=header_y_offset,
        )

        # draw children and costs
        edges_cost_tuples = [
            (e, c) for e, c in zip(node_info["edges"], node_info["costs"])
        ]
        edges_cost_tuples = sorted(edges_cost_tuples, key=lambda x: x[1])

        for idx, edges_cost_tuple in enumerate(edges_cost_tuples):
            self.simulation_canvas.fill_text(
                text=edges_cost_tuple[0][1],
                x=self.sim_canvas_size + 10,
                y=50 + idx * 16,
            )
            self.simulation_canvas.fill_text(
                text=round(edges_cost_tuple[1], 2),
                x=self.sim_canvas_size + 100,
                y=50 + idx * 16,
            )
        # draw explored
        for idx, explored in enumerate(simulation.explored_set):
            self.simulation_canvas.fill_text(
                text=explored,
                x=self.sim_canvas_size + 180,
                y=50 + idx * 16,
            )

        for idx, (frontier_item, avail, cost, heuristic, path_cost) in enumerate(
            zip(
                simulation.frontier_history["item"],
                simulation.frontier_history["avail"],
                simulation.frontier_history["cost"],
                simulation.frontier_history["heuristic"],
                simulation.frontier_history["path_cost"],
            )
        ):
            # frontier
            frontier_item_text = f"{frontier_item[0]}, {frontier_item[1]}"
            if idx == 0:
                header = "frontier"
            else:
                header = None
            self.draw_frontier_info(
                text=frontier_item_text,
                y_offset=50 + idx * (font_size + 1),
                x_offset=self.sim_canvas_size + 280,
                avail=avail,
                font_size=font_size,
                header=header,
                header_y_offset=header_y_offset,
            )

            # Dont draw cost and heuristic for BFS and DFS
            if simulation.search_method == "BFS" or simulation.search_method == "DFS":
                continue

            # cost
            if idx == 0:
                header = "cost f(n)"
            else:
                header = None
            self.draw_frontier_info(
                text=f"{cost:.2f}",
                y_offset=50 + idx * (font_size + 1),
                x_offset=self.sim_canvas_size + 390,
                avail=avail,
                font_size=font_size,
                header=header,
                header_y_offset=header_y_offset,
            )

            # path_cost
            if idx == 0:
                header = "path cost g(n)"
            else:
                header = None
            self.draw_frontier_info(
                text=f"{path_cost:.2f}",
                y_offset=50 + idx * (font_size + 1),
                x_offset=self.sim_canvas_size + 450,
                avail=avail,
                font_size=font_size,
                header=header,
                header_y_offset=header_y_offset,
            )

            # Dont draw heuristic for UCS
            if simulation.search_method == "UCS":
                continue

            # heuristic
            if idx == 0:
                header = "heuristic h(n)"
            else:
                header = None
            self.draw_frontier_info(
                text=f"{heuristic:.2f}",
                y_offset=50 + idx * (font_size + 1),
                x_offset=self.sim_canvas_size + 550,
                avail=avail,
                font_size=font_size,
                header=header,
                header_y_offset=header_y_offset,
            )

    def write_node_info_table2(self, simulation: Simulation, current_node: str):
        if current_node is None:
            return

        # set draw parameters and write head
        header_y_offset = 400
        self.simulation_canvas.fill_style = "#000000"
        font_size = 15
        self.simulation_canvas.font = "20px Arial"
        self.simulation_canvas.stroke_style = "#000000"
        self.simulation_canvas.line_width = 1
        self.simulation_canvas.fill_text(
            text="Frontier Dokumentation",
            x=self.sim_canvas_size + 100,
            y=header_y_offset - 25,
        )
        self.simulation_canvas.font = f"{font_size}px Arial"

        node_info = self.get_node_info(simulation=simulation, node=current_node)

        for idx, (frontier_item, step, cost, heuristic, path_cost) in enumerate(
            zip(
                simulation.frontier_history_unsorted["item"],
                simulation.frontier_history_unsorted["step"],
                simulation.frontier_history_unsorted["cost"],
                simulation.frontier_history_unsorted["heuristic"],
                simulation.frontier_history_unsorted["path_cost"],
            )
        ):
            # Vorgängerknoten
            frontier_item_text = f"{frontier_item[0]}"
            if idx == 0:
                header = "Vorgängerknoten"
            else:
                header = None
            self.draw_frontier_info(
                text=frontier_item_text,
                y_offset=header_y_offset + 50 + idx * (font_size + 1),
                x_offset=self.sim_canvas_size + 100,
                avail=True,
                font_size=font_size,
                header=header,
                header_y_offset=header_y_offset,
            )

            # Knotenname
            frontier_item_text = f"{frontier_item[1]}"
            if idx == 0:
                header = "Knotenname"
            else:
                header = None
            self.draw_frontier_info(
                text=frontier_item_text,
                y_offset=header_y_offset + 50 + idx * (font_size + 1),
                x_offset=self.sim_canvas_size + 10,
                avail=True,
                font_size=font_size,
                header=header,
                header_y_offset=header_y_offset,
            )

            # cost
            if idx == 0:
                header = "Kosten"
            else:
                header = None
            self.draw_frontier_info(
                text=f"{cost:.2f}",
                y_offset=header_y_offset + 50 + idx * (font_size + 1),
                x_offset=self.sim_canvas_size + 220,
                avail=True,
                font_size=font_size,
                header=header,
                header_y_offset=header_y_offset,
            )

            # Expansionsschritt
            frontier_item_text = f"{step}"
            if idx == 0:
                header = "Expansionsschritt"
            else:
                header = None
            self.draw_frontier_info(
                text=frontier_item_text,
                y_offset=header_y_offset + 50 + idx * (font_size + 1),
                x_offset=self.sim_canvas_size + 320,
                avail=True,
                font_size=font_size,
                header=header,
                header_y_offset=header_y_offset,
            )

    def draw_frontier_info(
        self,
        text: str,
        y_offset: int,
        x_offset: int,
        avail: bool,
        font_size: int,
        header: str = None,
        header_y_offset: int = None,
    ):
        if header is not None and header_y_offset is not None:
            self.simulation_canvas.fill_text(
                text=header,
                x=x_offset,
                y=header_y_offset,
            )

        self.simulation_canvas.fill_text(
            text=text,
            x=x_offset,
            y=y_offset,
        )
        if not avail:
            self.simulation_canvas.stroke_line(
                x1=x_offset,
                y1=y_offset - font_size / 2,
                x2=x_offset + len(text) * font_size / 2,
                y2=y_offset - font_size / 2,
            )

    def draw_nodes(self, simulation: Simulation, current_node: str):
        font_size = 15
        radius = int(font_size / 2) + 2
        self.simulation_canvas.text_align = "center"
        for node in list(simulation.sim_graph.nodes()):
            pos = simulation.sim_graph.nodes()[node]["pos"]
            scaled_pos_x, scaled_pos_y = self._rescale_to_simulation_frame_coords(
                np.array((pos[0], pos[1]))
            )
            # set color
            if node == current_node:
                self.simulation_canvas.fill_style = CURRENT_COLOR
            elif node == simulation.goal_node:
                self.simulation_canvas.fill_style = GOAL_COLOR
            elif node in simulation.explored_set:
                self.simulation_canvas.fill_style = EXPLORED_COLOR
            elif node in simulation.frontier_set:
                self.simulation_canvas.fill_style = IN_FRONTIER_COLOR
            else:
                self.simulation_canvas.fill_style = NORMAL_COLOR

            self.simulation_canvas.fill_circle(
                scaled_pos_x,
                scaled_pos_y,
                radius,
            )
            self.simulation_canvas.fill_style = "#000000"
            self.simulation_canvas.font = f"{font_size}px Arial"

            self.simulation_canvas.fill_text(
                text=node,
                x=scaled_pos_x,
                y=scaled_pos_y + int(font_size / 2) - 1,
            )
        self.simulation_canvas.text_align = "start"

    @staticmethod
    def circle_line_segment_intersection(
        circle_center, circle_radius, pt1, pt2, full_line=True, tangent_tol=1e-9
    ):
        """
        to avoid new container for shapely
        https://stackoverflow.com/questions/30844482/what-is-most-efficient-way-to-find-the-intersection-of-a-line-and-a-circle-in-py
        Find the points at which a circle intersects a line-segment.  This can happen at 0, 1, or 2 points.

        :param circle_center: The (x, y) location of the circle center
        :param circle_radius: The radius of the circle
        :param pt1: The (x, y) location of the first point of the segment
        :param pt2: The (x, y) location of the second point of the segment
        :param full_line: True to find intersections along full line - not just in the segment.  False will just return intersections within the segment.
        :param tangent_tol: Numerical tolerance at which we decide the intersections are close enough to consider it a tangent
        :return Sequence[Tuple[float, float]]: A list of length 0, 1, or 2, where each element is a point at which the circle intercepts a line segment.

        Note: We follow: http://mathworld.wolfram.com/Circle-LineIntersection.html
        """

        (p1x, p1y), (p2x, p2y), (cx, cy) = pt1, pt2, circle_center
        (x1, y1), (x2, y2) = (p1x - cx, p1y - cy), (p2x - cx, p2y - cy)
        dx, dy = (x2 - x1), (y2 - y1)
        dr = (dx**2 + dy**2) ** 0.5
        big_d = x1 * y2 - x2 * y1
        discriminant = circle_radius**2 * dr**2 - big_d**2

        if discriminant < 0:  # No intersection between circle and line
            return []
        else:  # There may be 0, 1, or 2 intersections with the segment
            intersections = [
                (
                    cx
                    + (
                        big_d * dy
                        + sign * (-1 if dy < 0 else 1) * dx * discriminant**0.5
                    )
                    / dr**2,
                    cy + (-big_d * dx + sign * abs(dy) * discriminant**0.5) / dr**2,
                )
                for sign in ((1, -1) if dy < 0 else (-1, 1))
            ]
            # This makes sure the order along the segment is correct
            if not full_line:
                # If only considering the segment, filter out intersections that do not fall within the segment
                fraction_along_segment = [
                    (xi - p1x) / dx if abs(dx) > abs(dy) else (yi - p1y) / dy
                    for xi, yi in intersections
                ]
                intersections = [
                    pt
                    for pt, frac in zip(intersections, fraction_along_segment)
                    if 0 <= frac <= 1
                ]
            if len(intersections) == 2 and abs(discriminant) <= tangent_tol:
                # If line is tangent to circle, return just one point (as both intersections have same location)
                return [intersections[0]]
            else:
                return intersections

    def get_endpoints(
        self, radius: int, p1: tuple[int, int], p2: tuple[int, int]
    ) -> list[tuple[int, int]]:

        ep1 = self.circle_line_segment_intersection(
            circle_center=p1, circle_radius=radius, pt1=p1, pt2=p2, full_line=False
        )

        ep2 = self.circle_line_segment_intersection(
            circle_center=p2, circle_radius=radius, pt1=p1, pt2=p2, full_line=False
        )
        return [ep1[0], ep2[0]]

    def get_midpoints(
        self, radius: int, p1: tuple[int, int], p2: tuple[int, int]
    ) -> tuple[int, int]:
        mid = (int(p1[0] + p2[0]) / 2, int(p1[1] + p2[1]) / 2)
        mid_1 = self.circle_line_segment_intersection(
            circle_center=mid, circle_radius=radius, pt1=p1, pt2=mid, full_line=False
        )
        mid_2 = self.circle_line_segment_intersection(
            circle_center=mid, circle_radius=radius, pt1=mid, pt2=p2, full_line=False
        )
        return [mid_1[0], mid_2[0]]

    def draw_edge(self, radius, p1, p2):
        end_node_1, end_node_2 = self.get_endpoints(radius=radius, p1=p1, p2=p2)
        mid1, mid2 = self.get_midpoints(radius=radius, p1=end_node_1, p2=end_node_2)

        self.simulation_canvas.begin_path()
        self.simulation_canvas.move_to(*end_node_1)
        self.simulation_canvas.line_to(*mid1)
        self.simulation_canvas.move_to(*mid2)
        self.simulation_canvas.line_to(*end_node_2)
        self.simulation_canvas.stroke()

    def draw_edges(self, simulation: Simulation):
        font_size = 15
        radius = int(font_size / 2) + 2
        self.simulation_canvas.line_width = 1
        all_edges = list(simulation.sim_graph.edges())
        self.simulation_canvas.stroke_style = "#000000"
        self.simulation_canvas.font = f"{font_size}px Arial"
        self.simulation_canvas.text_align = "center"
        # edge_cost colour
        self.simulation_canvas.fill_style = "#000000"
        self.simulation_canvas.stroke_style = "#000000"
        for edge in all_edges:
            pos1 = simulation.sim_graph.nodes()[edge[0]]["pos"]
            rescaled_node_1 = self._rescale_to_simulation_frame_coords(
                np.array((pos1[0], pos1[1]))
            )
            pos2 = simulation.sim_graph.nodes()[edge[1]]["pos"]
            rescaled_node_2 = self._rescale_to_simulation_frame_coords(
                np.array((pos2[0], pos2[1]))
            )
            self.draw_edge(radius=radius, p1=rescaled_node_1, p2=rescaled_node_2)
            scaled_pos_x = (rescaled_node_1[0] + rescaled_node_2[0]) / 2
            scaled_pos_y = (rescaled_node_1[1] + rescaled_node_2[1]) / 2

            self.simulation_canvas.fill_style = "#000000"
            self.simulation_canvas.fill_text(
                text=simulation.sim_graph.edges[edge]["dist"],
                x=scaled_pos_x,
                y=scaled_pos_y + int(font_size / 2) - 1,
            )
        self.simulation_canvas.text_align = "start"

    @staticmethod
    def rgb2hex(r: int, g: int, b: int):
        return "#{:02x}{:02x}{:02x}".format(r, g, b)

    def draw_available_path(self, simulation: Simulation, current_node: str):
        font_size = 15
        radius = int(font_size / 2) + 2
        self.simulation_canvas.line_width = 3
        edges_at_current_node = list(simulation.sim_graph.edges(current_node))
        for edge in edges_at_current_node:

            self.simulation_canvas.stroke_style = self.rgb2hex(0, 200, 0)
            pos = simulation.sim_graph.nodes()[edge[0]]["pos"]
            rescaled_node_1 = self._rescale_to_simulation_frame_coords(
                np.array((pos[0], pos[1]))
            )
            pos = simulation.sim_graph.nodes()[edge[1]]["pos"]
            rescaled_node_2 = self._rescale_to_simulation_frame_coords(
                np.array((pos[0], pos[1]))
            )
            self.draw_edge(radius=radius, p1=rescaled_node_1, p2=rescaled_node_2)

    def draw_current_path(self, simulation: Simulation):
        font_size = 15
        radius = int(font_size / 2) + 2
        self.simulation_canvas.stroke_style = "#FF0000"
        self.simulation_canvas.line_width = 3
        for edge in simulation.path:
            pos = simulation.sim_graph.nodes()[edge[0]]["pos"]
            rescaled_node_1 = self._rescale_to_simulation_frame_coords(
                np.array((pos[0], pos[1]))
            )
            pos = simulation.sim_graph.nodes()[edge[1]]["pos"]
            rescaled_node_2 = self._rescale_to_simulation_frame_coords(
                np.array((pos[0], pos[1]))
            )
            self.draw_edge(radius=radius, p1=rescaled_node_1, p2=rescaled_node_2)

    def draw_search_frame(
        self, simulation: Simulation, current_node: str, expand: bool = True
    ):
        self.simulation_canvas.clear_rect(
            0, 0, self.sim_canvas_size, self.sim_canvas_size
        )
        self.draw_nodes(simulation=simulation, current_node=current_node)
        self.draw_edges(simulation=simulation)
        self.write_node_info_table(simulation=simulation, current_node=current_node)
        self.write_node_info_table2(simulation=simulation, current_node=current_node)
        self.draw_current_path(simulation=simulation)

        if expand:
            self.draw_available_path(simulation=simulation, current_node=current_node)


class FieldVisualizer(Visualizer):
    def __init__(self, simulation: FieldSimulation, canvas_size: int = 400):
        self.sim_canvas_size = canvas_size
        self.simulation_canvas = Canvas(
            height=self.sim_canvas_size, width=int(self.sim_canvas_size * 16 / 9)
        )
        self._calculate_scaling_params_for_simulation(simulation=simulation)

    def _calculate_scaling_params_for_simulation(self, simulation: FieldSimulation):
        field_border_x = np.array(simulation.field_border[0])
        field_border_y = np.array(simulation.field_border[1])

        min_x = field_border_x.min()
        max_x = field_border_x.max()
        min_y = field_border_y.min()
        max_y = field_border_y.max()

        delta_x = max_x - min_x
        delta_y = max_y - min_y

        if delta_x > delta_y:
            high_delta = delta_x
        else:
            high_delta = delta_y

        self.sim_canvas_factor = self.sim_canvas_size / high_delta
        self.sim_canvas_x_offset = (
            abs(min_x) + (self.sim_canvas_size / self.sim_canvas_factor - delta_x) / 2
        )
        self.sim_canvas_y_offset = (
            abs(min_y) + (self.sim_canvas_size / self.sim_canvas_factor - delta_y) / 2
        )

        scaled_field_border_x = (
            field_border_x + self.sim_canvas_x_offset
        ) * self.sim_canvas_factor
        scaled_field_border_y = (
            field_border_y + self.sim_canvas_y_offset
        ) * self.sim_canvas_factor
        self.scaled_field_border = [
            (fb_x, fby)
            for fb_x, fby in zip(scaled_field_border_x, scaled_field_border_y)
        ]

        self.rows = [
            e
            for e in simulation.sim_graph.edges()
            if simulation.sim_graph.edges[e]["headland"] == False
        ]

        x_row_coords = [
            np.array(
                (
                    simulation.sim_graph.nodes[a]["pos"][0],
                    simulation.sim_graph.nodes[b]["pos"][0],
                )
            )
            for a, b in self.rows
        ]
        y_row_coords = [
            np.array(
                (
                    simulation.sim_graph.nodes[a]["pos"][1],
                    simulation.sim_graph.nodes[b]["pos"][1],
                )
            )
            for a, b in self.rows
        ]
        row_coords = np.array((x_row_coords, y_row_coords))
        scaled_row_coords = self._rescale_to_simulation_frame_coords(row_coords)

        self.scaled_row_coords = [
            ([x[0], y[0]], [x[1], y[1]])
            for x, y in zip(scaled_row_coords[0], scaled_row_coords[1])
        ]

        logistic_nodes = []
        for names in simulation.logistic_point_names:
            logistic_nodes.append(names)

        self.scaled_logistic_coords = []
        for logistic_node in logistic_nodes:
            logistic_pos = np.array(simulation.sim_graph.nodes[logistic_node]["pos"])
            logistic_pos = self._rescale_to_simulation_frame_coords(logistic_pos)

    def get_node_info(self, simulation: FieldSimulation, node: str) -> dict:
        edges = [
            e
            for e in list(simulation.sim_graph.edges(node))
            if simulation.sim_graph.edges[e]["headland"] == True
        ]
        costs = [simulation.sim_graph.edges[e]["dist"] for e in edges]

        info = {
            "node": node,
            "edges": edges,
            "costs": costs,
        }
        return info

    def draw_borders(self):
        self.simulation_canvas.stroke_style = "#000000"
        self.simulation_canvas.line_width = 1
        self.simulation_canvas.stroke_polygon(self.scaled_field_border)
        self.simulation_canvas.fill_style = "#997950"
        self.simulation_canvas.fill_polygon(self.scaled_field_border)

    def draw_nodes(self, simulation: FieldSimulation, current_node: str):
        radius = 8
        for node in list(simulation.sim_graph.nodes()):
            pos = simulation.sim_graph.nodes()[node]["pos"]
            scaled_pos_x, scaled_pos_y = self._rescale_to_simulation_frame_coords(
                np.array((pos[0], pos[1]))
            )
            # set color
            if node == current_node:
                self.simulation_canvas.fill_style = CURRENT_COLOR
            elif "logistic" in node:
                self.simulation_canvas.fill_style = GOAL_COLOR
            elif node in simulation.explored_set:
                self.simulation_canvas.fill_style = EXPLORED_COLOR
            elif node in simulation.frontier_set:
                self.simulation_canvas.fill_style = IN_FRONTIER_COLOR
            else:
                self.simulation_canvas.fill_style = NORMAL_COLOR

            self.simulation_canvas.fill_circle(
                scaled_pos_x,
                scaled_pos_y,
                radius,
            )
            self.simulation_canvas.fill_style = "#000000"
            self.simulation_canvas.font = "15px Arial"

            self.simulation_canvas.fill_text(
                text=node,
                x=scaled_pos_x,
                y=scaled_pos_y,
            )

    def draw_rows(self, simulation: FieldSimulation):
        row_width = (
            simulation.field_info_dict["operating_width"] * self.sim_canvas_factor + 1
        )

        self.simulation_canvas.stroke_style = "#4c392d"
        self.simulation_canvas.line_width = row_width
        for row_coord in self.scaled_row_coords:
            self.simulation_canvas.stroke_line(*row_coord[0], *row_coord[1])

    @staticmethod
    def rgb2hex(r: int, g: int, b: int):
        return "#{:02x}{:02x}{:02x}".format(r, g, b)

    def draw_available_path(self, simulation: FieldSimulation, current_node: str):
        self.simulation_canvas.line_width = 3
        edges_at_current_node = list(simulation.headland_subgraph.edges(current_node))
        for idx, edge in enumerate(edges_at_current_node):
            r = int(idx / len(edges_at_current_node) * 200)
            g = 200 - int(idx / len(edges_at_current_node) * 200)
            self.simulation_canvas.stroke_style = self.rgb2hex(r, g, 140)
            dubin_x = simulation.headland_subgraph.edges()[edge]["path_x"]
            dubin_y = simulation.headland_subgraph.edges()[edge]["path_y"]
            rescaled_path = self._rescale_to_simulation_frame_coords(
                np.array([dubin_x, dubin_y])
            )
            lines = [
                np.array((x, y)) for x, y in zip(rescaled_path[0], rescaled_path[1])
            ]
            self.simulation_canvas.begin_path()
            self.simulation_canvas.move_to(*lines[0])
            for p in lines:
                self.simulation_canvas.line_to(*p)
            self.simulation_canvas.stroke()

    def draw_current_path(self, simulation: FieldSimulation):
        self.simulation_canvas.stroke_style = "#000000"
        for edge in simulation.path:
            dubin_x = simulation.sim_graph.edges()[edge]["path_x"]
            dubin_y = simulation.sim_graph.edges()[edge]["path_y"]
            rescaled_path = self._rescale_to_simulation_frame_coords(
                np.array([dubin_x, dubin_y])
            )
            lines = [
                np.array((x, y)) for x, y in zip(rescaled_path[0], rescaled_path[1])
            ]
            self.simulation_canvas.begin_path()
            self.simulation_canvas.move_to(*lines[0])
            for p in lines:
                self.simulation_canvas.line_to(*p)
            self.simulation_canvas.stroke()

    def draw_search_frame(
        self, simulation: FieldSimulation, current_node: str, expand: bool = True
    ):
        self.simulation_canvas.clear_rect(
            0, 0, self.sim_canvas_size, self.sim_canvas_size
        )

        self.draw_borders()
        self.draw_rows(simulation=simulation)
        self.draw_nodes(simulation=simulation, current_node=current_node)
        self.write_node_info_table(simulation=simulation, current_node=current_node)
        self.write_node_info_table2(simulation=simulation, current_node=current_node)
        self.draw_current_path(simulation=simulation)

        if expand:
            self.draw_available_path(simulation=simulation, current_node=current_node)


class IPyUI:
    def __init__(self, field_vis: Visualizer, search_sim: Simulation):
        self.field_vis = field_vis
        self.search_sim = search_sim

    def search_ui(self):
        # Select and expand node
        self.node_names = natsorted(list(self.search_sim.sim_graph.nodes()))
        self.selected_node = self.node_names[0]
        self.node_dd = widgets.Dropdown(
            options=[*self.node_names],
            value=self.selected_node,
            description="Initial Node",
            disabled=False,
        )
        self.node_dd.observe(self._node_dd_callback, names="value")
        self.initial_node_btnn = widgets.Button(description="Choose initial node")
        self.initial_node_btnn.on_click(self._initial_node_btnn_callback)
        expand_box = widgets.HBox([self.node_dd, self.initial_node_btnn])

        # Search settings
        self.search_type = "graph"
        self.tree_vs_graph_dd = widgets.Dropdown(
            options=["tree", "graph"],
            value=self.search_type,
            description="S. Type",
            disabled=True,
        )

        self.search_method = "BFS"
        self.search_dd = widgets.Dropdown(
            options=["BFS", "DFS", "UCS", "GreedyBestFirst", "A*"],
            value=self.search_method,
            description="S. Method",
            disabled=False,
        )
        settings_box = widgets.HBox([self.tree_vs_graph_dd, self.search_dd])

        self.choose_btnn = widgets.Button(description="Choose next (POP)")
        self.choose_btnn.disabled = True
        self.goal_test_btnn = widgets.Button(description="Goal test")
        self.goal_test_btnn.disabled = True
        self.expand_btnn = widgets.Button(description="Expand chosen")
        self.expand_btnn.disabled = True
        search_box = widgets.HBox(
            [self.choose_btnn, self.goal_test_btnn, self.expand_btnn]
        )
        self.choose_btnn.on_click(self._choose)
        self.goal_test_btnn.on_click(self._goal_test)
        self.expand_btnn.on_click(self._expand)
        self.search_dd.observe(self._search_method_dd_callback, names="value")
        self.tree_vs_graph_dd.observe(self._search_type_dd_callback, names="value")

        items = [
            settings_box,
            expand_box,
            search_box,
        ]
        display(widgets.VBox(items), self.field_vis.simulation_canvas)

        self.field_vis.draw_search_frame(
            self.search_sim, current_node=None, expand=False
        )

    def _initial_node_btnn_callback(self, _):
        # init search
        self.search_sim.init_search(
            start_node=self.selected_node,
            search_type=self.search_type,
            search_method=self.search_method,
        )
        # set button availibility
        self.tree_vs_graph_dd.disabled = True
        self.search_dd.disabled = True
        self.node_dd.disabled = True
        self.choose_btnn.disabled = True
        self.goal_test_btnn.disabled = False
        self.initial_node_btnn.disabled = True
        # visualize
        self.field_vis.draw_search_frame(
            self.search_sim, current_node=self.selected_node, expand=False
        )

    def _node_dd_callback(self, msg):
        self.selected_node = msg["new"]

    def _search_method_dd_callback(self, msg):
        self.search_method = msg["new"]

    def _search_type_dd_callback(self, msg):
        self.search_type = msg["new"]

    def _choose(self, _):
        self.selected_node = self.search_sim.choose()
        self.search_sim.create_current_path(last_node=self.selected_node)
        self.choose_btnn.disabled = True
        # DFS and BFS are not using goal test button (included in expand)
        if (
            self.search_sim.search_method != "DFS"
            and self.search_sim.search_method != "BFS"
        ):
            self.goal_test_btnn.disabled = False
            self.expand_btnn.disabled = True
        else:
            self.expand_btnn.disabled = False
        self.field_vis.draw_search_frame(
            self.search_sim, current_node=self.selected_node, expand=False
        )

    def _expand(self, _):
        # DFS and BFS have goal_test included in expand to always check frontier
        goal_reached, goal_node = self.search_sim.expand()
        if goal_reached:
            self.handle_goal_reached(goal_node=goal_node)
        else:
            self.choose_btnn.disabled = False
            self.goal_test_btnn.disabled = True
            self.expand_btnn.disabled = True
            self.field_vis.draw_search_frame(
                self.search_sim, current_node=self.selected_node
            )

    def _goal_test(self, _):
        goal_reached = self.search_sim.goal_test(node=self.selected_node)
        if goal_reached:
            self.handle_goal_reached(goal_node=self.selected_node)
        else:
            self.choose_btnn.disabled = True
            self.goal_test_btnn.disabled = True
            self.expand_btnn.disabled = False

    def handle_goal_reached(self, goal_node: str):
        self.choose_btnn.disabled = True
        self.expand_btnn.disabled = True
        self.goal_test_btnn.disabled = True
        self.search_sim.chosen = goal_node
        self.search_sim.create_current_path(last_node=goal_node)
        self.field_vis.draw_search_frame(
            self.search_sim, current_node=goal_node, expand=False
        )
        print(f"Found target {goal_node}")
        print(f"Path cost = {self.search_sim.get_cost_of_current_path():.2f}")
