# adapted from https://gymnasium.farama.org/tutorials/training_agents/blackjack_tutorial/
# Author: Till Zemann
# License: MIT License
# By Jonas Boysen

import matplotlib.pyplot as plt
import numpy as np
import math
import ipywidgets as widgets
import gymnasium as gym
from IPython.display import clear_output
from tqdm import tqdm
from ipycanvas import Canvas
import matplotlib.pyplot as plt
from IPython.display import clear_output
from tqdm import tqdm
from collections import defaultdict
from PIL import ImageColor
from IPython.display import display


class CliffAgent:
    def __init__(
        self,
        learning_rate: float,
        initial_epsilon: float,
        final_epsilon: float,
        discount_factor: float = 0.95,
        show_env: bool = False,
        max_steps_per_epoch: int = 1000,
        update_algorithm: str = "q_learning",
    ):
        """Initialize a Reinforcement Learning agent with an empty dictionary
        of state-action values (q_values), a learning rate and an epsilon.

        Args:
            learning_rate: The learning rate
            initial_epsilon: The initial epsilon value
            epsilon_decay: The decay for epsilon
            final_epsilon: The final epsilon value
            discount_factor: The discount factor for computing the Q-value
        """
        self.env = gym.make("CliffWalking-v0", render_mode="rgb_array")
        self.obs = None
        self.last_reward = None
        self.last_obs = None
        self.last_action = None
        self.initial_tile = 36
        self.steps = 0
        self.reset_env()

        self.q_values = defaultdict(lambda: np.zeros(self.env.action_space.n))
        self.done = False
        self.show_env = show_env

        self.lr = learning_rate
        self.discount_factor = discount_factor

        self.epsilon = initial_epsilon
        self.initial_epsilon = initial_epsilon
        self.final_epsilon = final_epsilon
        self.update_algorithm = update_algorithm
        self.max_steps_per_epoch = max_steps_per_epoch

        self.training_error = []
        self.rewards = []
        self.epsilons = []
        self.episode_lengths = []

    def reset_env(self):
        self.obs, info = self.env.reset()
        self.done = False
        self.steps = 0
        # to skip first update in SARSA (waiting for selected next action)
        self.last_reward = None
        self.last_obs = None
        self.last_action = None

    def get_action(self, obs: tuple[int, int, bool]) -> int:
        """
        Returns the best action with probability (1 - epsilon)
        otherwise a random action with probability epsilon to ensure exploration.
        """
        # with probability epsilon return a random action to explore the environment
        if np.random.random() < self.epsilon:
            return self.env.action_space.sample()

        # with probability (1 - epsilon) act greedily (exploit)
        else:
            return int(np.argmax(self.q_values[obs]))

    def update(
        self,
        obs: tuple[int, int, bool],
        action: int,
        reward: float,
        terminated: bool,
        next_obs: tuple[int, int, bool],
        print_info: bool = False,
        algorithm: str = "q_learning",
    ):
        """Updates the Q-value of an action."""

        # update the current q_value using Q-Learning
        if algorithm == "q_learning":
            self._q_learning_update(
                obs=obs,
                action=action,
                reward=reward,
                terminated=terminated,
                next_obs=next_obs,
                print_info=print_info,
            )

        # update the last q_value using SARSA
        elif algorithm == "sarsa":
            self._sarsa_update(
                obs=obs,
                action=action,
                reward=reward,
                terminated=terminated,
                next_obs=next_obs,
                print_info=print_info,
            )
            # update last step with Q(s',a') = 0
            if terminated:
                # Q(s',a')
                future_q_value = 0
                # TD = r + gamma * Q(s',a') - Q(s,a)
                temporal_difference = (
                    reward
                    + self.discount_factor * future_q_value
                    - self.q_values[obs][action]
                )
                # Q(s,a)_new = Q(s,a) + alpha * TD
                self.q_values[obs][action] = (
                    self.q_values[obs][action] + self.lr * temporal_difference
                )
                self.training_error.append(temporal_difference)

    def _q_learning_update(
        self,
        obs: tuple[int, int, bool],
        action: int,
        reward: float,
        terminated: bool,
        next_obs: tuple[int, int, bool],
        print_info: bool = False,
    ):
        # max(a') Q(s',a')
        future_q_value = (not terminated) * np.max(self.q_values[next_obs])
        # TD = r + gamma * max(a') Q(s',a') - Q(s,a)
        temporal_difference = (
            reward + self.discount_factor * future_q_value - self.q_values[obs][action]
        )
        # Q(s,a)_new = Q(s,a) + alpha * TD
        self.q_values[obs][action] = (
            self.q_values[obs][action] + self.lr * temporal_difference
        )
        self.training_error.append(temporal_difference)

        if print_info:
            self.print_update_info(
                obs=next_obs,
                action=action,
                reward=reward,
                terminated=terminated,
                next_obs=next_obs,
                temporal_difference=temporal_difference,
                future_q_value=future_q_value,
            )

    def _sarsa_update(
        self,
        obs: tuple[int, int, bool],
        action: int,
        reward: float,
        terminated: bool,
        next_obs: tuple[int, int, bool],
        print_info: bool = True,
    ):
        # skip first update
        if (
            self.last_reward is None
            or self.last_obs is None
            or self.last_action is None
        ):
            pass
        else:
            # Q(s',a')
            future_q_value = self.q_values[obs][action]
            # TD = r + gamma * Q(s',a') - Q(s,a)
            temporal_difference = (
                self.last_reward
                + self.discount_factor * future_q_value
                - self.q_values[self.last_obs][self.last_action]
            )
            # Q(s,a)_new = Q(s,a) + alpha * TD
            self.q_values[self.last_obs][self.last_action] = (
                self.q_values[self.last_obs][self.last_action]
                + self.lr * temporal_difference
            )

            self.training_error.append(temporal_difference)

            if print_info:
                self.print_update_info(
                    obs=obs,
                    action=action,
                    reward=reward,
                    terminated=terminated,
                    next_obs=next_obs,
                    temporal_difference=temporal_difference,
                    future_q_value=future_q_value,
                )

        # buffer last values for sarsa algorithm
        self.last_reward = reward
        self.last_obs = obs
        self.last_action = action

    def decay_epsilon(self, decay_over_n_epochs: int):
        epsilon_decay = self.initial_epsilon / (decay_over_n_epochs / 2)
        self.epsilon = max(self.final_epsilon, self.epsilon - epsilon_decay)

    def perform_epoch(self, eps_decay_over_n_epochs: int):
        reward = 0
        steps = 0
        while not self.done and steps < self.max_steps_per_epoch:
            reward += self.perform_step(action_select="eps_decay")
            steps += 1

        self.rewards.append(reward)
        self.episode_lengths.append(self.steps)
        self.epsilons.append(self.epsilon)
        self.decay_epsilon(decay_over_n_epochs=eps_decay_over_n_epochs)
        if self.done:
            self.reset_env()

    def perform_step(
        self, action_select: str = "eps_decay", print_info: bool = False
    ) -> int:
        if action_select == "eps_decay":
            action = self.get_action(self.obs)
        elif action_select == "up":
            action = 0
        elif action_select == "right":
            action = 1
        elif action_select == "down":
            action = 2
        elif action_select == "left":
            action = 3
        elif action_select == "greedy":
            action = int(np.argmax(self.q_values[self.obs]))

        next_obs, reward, terminated, truncated, info = self.env.step(action)

        # update the agent
        self.update(
            self.obs,
            action,
            reward,
            terminated,
            next_obs,
            print_info=print_info,
            algorithm=self.update_algorithm,
        )

        if self.show_env:
            self.render_frame()

        # update the current obs and check if episode has finished
        self.done = terminated or truncated
        self.obs = next_obs

        self.steps += 1
        return reward

    def print_update_info(
        self,
        obs: tuple[int, int, bool],
        action: int,
        reward: float,
        terminated: bool,
        next_obs: tuple[int, int, bool],
        temporal_difference: float,
        future_q_value: float,
    ):
        print(
            "\n_________________________________________________________________________________________________\n"
        )
        if self.update_algorithm == "q_learning":
            print("Update Q of current state with Q_learning:\n")
            print(
                f"Current state: {obs}\nAction: {action}\nReward: {reward}\nTerminated: {terminated}\nNext State: {next_obs}\n\n"
            )
            print(f"Q(s',a') for all a': {*self.q_values[next_obs],}\n")
            print(f"max(Q(s',a')): {future_q_value} \n")

            print("Calculate the temporal difference:")
            self.print_alligned(
                str1="temporal_difference = reward + discount_factor * max(Q(s',a')) - Q(s,a)",
                str2=f"{temporal_difference} = {reward} + {self.discount_factor} * {future_q_value} - {self.q_values[obs][action]}",
            )
            print("\nCalculate the new Q-value")
            self.print_alligned(
                str1="Q_new(s,a) = Q(s,a) + learning_rate * temporal_difference",
                str2=f"{self.q_values[obs][action]} = {self.q_values[obs][action]} + {self.lr} * {temporal_difference}",
            )
        if self.update_algorithm == "sarsa":
            print("Update Q of last state SARSA:\n")

            print("Viewing from last state:\n")
            print(
                f"Current state: {self.last_obs}\nAction: {self.last_action}\nReward: {self.last_reward}\nTerminated: {terminated}\nNext State: {obs}\n\n"
            )
            print(f"Q(s',a') for all a': {*self.q_values[self.last_obs],}\n")
            print(f"Q(s',a') with selected a': {future_q_value} \n")

            print("Calculate the temporal difference:")
            self.print_alligned(
                str1="temporal_difference = reward + discount_factor * max(Q(s',a')) - Q(s,a)",
                str2=f"{temporal_difference} = {self.last_reward} + {self.discount_factor} * {future_q_value} - {self.q_values[self.last_obs][self.last_action]}",
            )
            print("\nCalculate the new Q-value")
            self.print_alligned(
                str1="Q_new(s,a) = Q(s,a) + learning_rate * temporal_difference",
                str2=f"{self.q_values[self.last_obs][self.last_action]} = {self.q_values[self.last_obs][self.last_action]} + {self.lr} * {temporal_difference}",
            )

    def print_alligned(self, str1: str, str2: str, sep: str = " ") -> tuple[str, str]:
        str1_words = str1.split(sep=sep)
        str2_words = str2.split(sep=sep)
        assert len(str1_words) == len(str2_words)
        alligned_str1_words = []
        alligned_str2_words = []
        for w1, w2 in zip(str1_words, str2_words):
            alligned_w1, alligned_w2 = self.create_central_strings(w1, w2)
            alligned_str1_words.append(alligned_w1)
            alligned_str2_words.append(alligned_w2)

        str1 = " ".join(alligned_str1_words)
        str2 = " ".join(alligned_str2_words)
        print(f"{str1}\n{str2}")

    @staticmethod
    def create_central_strings(str1: str, str2: str) -> tuple[str, str]:
        if len(str1) == len(str2):
            return (str1, str2)

        if len(str1) > len(str2):
            str2 = str2.center(len(str1))
            return (str1, str2)

        if len(str1) < len(str2):
            str1 = str1.center(len(str2))
            return (str1, str2)

    def render_frame(self):
        clear_output(wait=True)
        plt.imshow(self.env.render())
        plt.show()

    def vis_results(self, n_episodes):
        divisor = 4
        if n_episodes <= divisor:
            return
        rolling_length = 10
        fig, axs = plt.subplots(nrows=2, ncols=3, figsize=(12, 5))
        axs[0][0].set_title("Episode rewards")
        # compute and assign a rolling average of the data to provide a smoother graph
        reward_moving_average = (
            np.convolve(
                np.array(self.rewards),
                np.ones(rolling_length),
                mode="same",
            )
            / rolling_length
        )
        axs[0][0].plot(range(len(reward_moving_average)), reward_moving_average)
        axs[0][1].plot(self.rewards)
        axs[0][1].set_ylim(-100, 0)
        axs[0][1].set_title("Episode rewards (zoomed)")

        axs[0][2].set_title("Episode lengths")
        length_moving_average = (
            np.convolve(
                np.array(self.episode_lengths),
                np.ones(rolling_length),
                mode="same",
            )
            / rolling_length
        )
        axs[0][2].plot(range(len(length_moving_average)), length_moving_average)

        axs[1][2].set_title("Episode lengths (zoomed)")
        axs[1][2].plot(self.episode_lengths)
        axs[1][2].set_ylim(0, 200)

        axs[1][0].set_title("Epsilon")
        axs[1][0].plot(range(len(self.epsilons)), self.epsilons)

        axs[1][1].set_title("Training Error")
        training_error_moving_average = (
            np.convolve(
                np.array(self.training_error), np.ones(rolling_length), mode="same"
            )
            / rolling_length
        )
        axs[1][1].plot(
            range(len(training_error_moving_average)), training_error_moving_average
        )
        plt.tight_layout()
        plt.show()


class CliffPolicyVisualizer:
    """
    Class to visualize the q table of cliff walker
    """

    def __init__(
        self,
        agent: CliffAgent,
        canvas_size: int,
        rows: int = 4,
        cols: int = 12,
        border: int = 10,
        n_episodes: int = 1,
        font_size: int = 16,
        all_white: bool = False,
    ):
        self.agent = agent

        self.sim_canvas_size = canvas_size
        self.rows = rows
        self.cols = cols
        self.border = border

        self.initial_state = 36

        self.simulation_canvas = Canvas(
            height=self.sim_canvas_size[1], width=self.sim_canvas_size[0]
        )

        self.perform_action_bttn = widgets.Button(description="Perform next action")
        self.perform_action_bttn.on_click(self.perform_step_callback)

        self.action_choices = ["up", "right", "down", "left", "greedy", "eps_decay"]
        self.selected_action = self.action_choices[0]
        self.action_dd = widgets.Dropdown(
            options=[*self.action_choices],
            value=self.selected_action,
            description="Action",
            disabled=False,
        )
        self.action_dd.observe(self._action_dd_callback, names="value")

        self.perform_epoch_bttn = widgets.Button(description="Perform epoch(s)")
        self.perform_epoch_bttn.on_click(self.perform_epoch_callback)

        self.n_epochs = n_episodes
        self.n_epochs_slider = widgets.IntSlider(
            value=self.n_epochs, min=1, max=n_episodes, step=1
        )
        self.n_epochs_slider.observe(self.n_epochs_slider_callback, names="value")

        self.animate_epoch = False
        self.animate_epoch_chkbx = widgets.Checkbox(
            value=self.animate_epoch, description="Animate epoch(s)"
        )
        self.animate_epoch_chkbx.observe(
            self.animate_epoch_chkbx_callback, names="value"
        )

        action_box = widgets.HBox([self.action_dd, self.perform_action_bttn])
        epoch_box = widgets.HBox(
            [self.n_epochs_slider, self.animate_epoch_chkbx, self.perform_epoch_bttn]
        )
        ui_box = widgets.VBox([action_box, epoch_box])
        frame = widgets.VBox([ui_box, self.simulation_canvas])

        self.last_tile_idx = 36
        self.penultimate_tile_idx = 36

        # set constants
        self.show_decimals = 2
        self.font_size = font_size
        self.grass_color = "#84eab3"
        self.cliff_color = "#d3d3d3"
        self.current_color = "#dd3333"
        self.arrow_color = "#000000"
        self.simulation_canvas.text_align = "center"
        self.simulation_canvas.text_baseline = "middle"
        self.simulation_canvas.stroke_style = "#000000"
        self.simulation_canvas.fill_style = "#000000"
        self.simulation_canvas.font = f"{self.font_size}px sans-serif"
        if all_white:
            self.simulation_canvas.fill_style = "#FFFFFF"
            self.grass_color = "#FFFFFF"
            self.cliff_color = "#FFFFFF"
            self.current_color = "#FFFFFF"
            self.arrow_color = "#FFFFFF"
        self.init_parameters()

        self.draw_grid_world(agent_pos=36)
        display(frame)

    def init_parameters(self):
        row_min_size = (self.sim_canvas_size[1] - 2 * self.border) / (self.rows + 1)
        col_min_size = (self.sim_canvas_size[0] - 2 * self.border) / (self.cols + 1)
        size = row_min_size if row_min_size < col_min_size else col_min_size
        self.size = math.floor(size)
        self.right_border = self.sim_canvas_size[0] - size * self.cols - self.border
        self.bottom_border = self.sim_canvas_size[1] - size * self.rows - self.border

        self.grid = list(range(self.rows * self.cols))
        for row in range(self.rows):
            for col in range(self.cols):
                tile_color = self.grass_color
                if row == self.rows - 1:
                    if not (col == 0 or col == self.cols - 1):
                        tile_color = self.cliff_color
                self.grid[col + self.cols * row] = {
                    "x": self.border + col * self.size,
                    "y": self.border + row * self.size,
                    "color": tile_color,
                    "q_values": np.zeros(4, dtype=np.float32),
                }

        # create tile coords and colors once for fast drawing
        self.tiles_x = [tile["x"] for tile in self.grid]
        self.tiles_y = [tile["y"] for tile in self.grid]
        self.tiles_color = [self.hex2rgb(tile["color"]) for tile in self.grid]

        # create arrow shapes once for fast drawing
        rect_size = int(self.size / 8)
        self.arrows = []
        for tile in self.grid:
            arrow_up = np.array(
                [
                    (
                        tile["x"] + int(self.size / 2) - rect_size,
                        tile["y"] + int(self.size / 2),
                    ),
                    (
                        tile["x"] + int(self.size / 2) + rect_size,
                        tile["y"] + int(self.size / 2),
                    ),
                    (
                        tile["x"] + int(self.size / 2),
                        tile["y"] + int(self.size / 2) - rect_size,
                    ),
                ]
            )

            arrow_right = np.array(
                [
                    (
                        tile["x"] + int(self.size / 2),
                        tile["y"] + int(self.size / 2) - rect_size,
                    ),
                    (
                        tile["x"] + int(self.size / 2),
                        tile["y"] + int(self.size / 2) + rect_size,
                    ),
                    (
                        tile["x"] + int(self.size / 2) + rect_size,
                        tile["y"] + int(self.size / 2),
                    ),
                ]
            )
            arrow_down = np.array(
                [
                    (
                        tile["x"] + int(self.size / 2) - rect_size,
                        tile["y"] + int(self.size / 2),
                    ),
                    (
                        tile["x"] + int(self.size / 2) + rect_size,
                        tile["y"] + int(self.size / 2),
                    ),
                    (
                        tile["x"] + int(self.size / 2),
                        tile["y"] + int(self.size / 2) + rect_size,
                    ),
                ]
            )
            arrow_left = np.array(
                [
                    (
                        tile["x"] + int(self.size / 2),
                        tile["y"] + int(self.size / 2) - rect_size,
                    ),
                    (
                        tile["x"] + int(self.size / 2),
                        tile["y"] + int(self.size / 2) + rect_size,
                    ),
                    (
                        tile["x"] + int(self.size / 2) - rect_size,
                        tile["y"] + int(self.size / 2),
                    ),
                ]
            )
            self.arrows.append([arrow_up, arrow_right, arrow_down, arrow_left])

    @staticmethod
    def rgb2hex(r: int, g: int, b: int):
        return "#{:02x}{:02x}{:02x}".format(r, g, b)

    @staticmethod
    def hex2rgb(hex: str):
        return ImageColor.getcolor(hex, "RGB")

    def perform_step_callback(self, _=None):
        self.agent.perform_step(action_select=self.selected_action, print_info=True)
        if self.agent.done:
            self.agent.reset_env()
        self.update_grid(self.agent.q_values, self.agent.obs)

    def perform_epoch_callback(self, _=None):
        for _ in tqdm(range(self.n_epochs)):
            self.agent.perform_epoch(eps_decay_over_n_epochs=self.n_epochs)
            self.agent.reset_env()
            # if each epoch should be visualized
            if self.animate_epoch:
                self.draw_grid_world(self.agent.q_values, self.initial_state)
                self.update_grid(self.agent.q_values, self.initial_state)
        # otherwise visualize once at the end
        if not self.animate_epoch:
            self.draw_grid_world(self.agent.q_values, self.initial_state)
            self.update_grid(self.agent.q_values, self.initial_state)

        self.agent.vis_results(n_episodes=self.n_epochs)

    def animate_epoch_chkbx_callback(self, msg):
        self.animate_epoch = msg["new"]

    def _action_dd_callback(self, msg):
        self.selected_action = msg["new"]

    def _action_dd_callback(self, msg):
        self.selected_action = msg["new"]

    def n_epochs_slider_callback(self, msg):
        self.n_epochs = msg["new"]

    def update_grid(self, q_values: dict, obs: int):
        # update q_values
        self.grid[obs]["q_values"] = q_values[obs]
        self.grid[self.last_tile_idx]["q_values"] = q_values[self.last_tile_idx]
        self.grid[self.penultimate_tile_idx]["q_values"] = q_values[
            self.penultimate_tile_idx
        ]

        # refresh last tile with orig color
        arg_max_q_last = self.grid[self.last_tile_idx]["q_values"].argmax()
        self.refresh_tile(
            tile_idx=self.last_tile_idx,
            color=self.grid[self.last_tile_idx]["color"],
            arrow_direction=arg_max_q_last,
        )
        # refresh current tile with current color
        arg_max_q_current = self.grid[obs]["q_values"].argmax()
        self.refresh_tile(
            tile_idx=obs,
            color=self.current_color,
            arrow_direction=arg_max_q_current,
        )
        # if sarsa is selected also refresh the penultimate
        # since these q values are updated
        if self.agent.update_algorithm == "sarsa":
            arg_max_q_penultimate = self.grid[self.penultimate_tile_idx][
                "q_values"
            ].argmax()
            if self.penultimate_tile_idx == obs:
                color = self.current_color
            else:
                color = self.grid[self.penultimate_tile_idx]["color"]
            self.refresh_tile(
                tile_idx=self.penultimate_tile_idx,
                color=color,
                arrow_direction=arg_max_q_penultimate,
            )
        # set last tile idx as punultimate tile idx (for SARSA algorithm)
        self.penultimate_tile_idx = self.last_tile_idx
        # set current idx as next last tile idx
        self.last_tile_idx = obs

    def refresh_tile(self, tile_idx: int, color: str, arrow_direction: int):
        self.simulation_canvas.fill_style = color
        self.simulation_canvas.fill_rect(
            x=self.grid[tile_idx]["x"],
            y=self.grid[tile_idx]["y"],
            width=self.size,
            height=self.size,
        )
        self.simulation_canvas.stroke_rect(
            x=self.grid[tile_idx]["x"],
            y=self.grid[tile_idx]["y"],
            width=self.size,
            height=self.size,
        )

        self.simulation_canvas.fill_style = self.arrow_color
        self.simulation_canvas.fill_polygon(self.arrows[tile_idx][arrow_direction])
        # move up
        self.simulation_canvas.fill_text(
            x=self.grid[tile_idx]["x"] + int(self.size / 2),
            y=self.grid[tile_idx]["y"] + self.font_size,
            text=f'{self.grid[tile_idx]["q_values"][0]:.{self.show_decimals}f}',
        )
        # q move right
        self.simulation_canvas.fill_text(
            x=self.grid[tile_idx]["x"]
            + self.size
            - self.font_size * self.show_decimals * 0.75,
            y=self.grid[tile_idx]["y"] + int(self.size / 2),
            text=f'{self.grid[tile_idx]["q_values"][1]:.{self.show_decimals}f}',
        )
        # q move down
        self.simulation_canvas.fill_text(
            x=self.grid[tile_idx]["x"] + int(self.size / 2),
            y=self.grid[tile_idx]["y"] + self.size - self.font_size,
            text=f'{self.grid[tile_idx]["q_values"][2]:.{self.show_decimals}f}',
        )
        # q move left
        self.simulation_canvas.fill_text(
            x=self.grid[tile_idx]["x"] + self.font_size * self.show_decimals * 0.75,
            y=self.grid[tile_idx]["y"] + int(self.size / 2),
            text=f'{self.grid[tile_idx]["q_values"][3]:.{self.show_decimals}f}',
        )

    def draw_grid_world(self, q_values=None, agent_pos=None):
        if q_values is not None:
            for idx in range(len(self.grid)):
                self.grid[idx]["q_values"] = q_values[idx]
        # draw environment
        colors = self.tiles_color.copy()
        if agent_pos is not None:
            colors[agent_pos] = self.hex2rgb(self.current_color)

        self.simulation_canvas.fill_styled_rects(
            x=self.tiles_x,
            y=self.tiles_y,
            width=self.size,
            height=self.size,
            color=colors,
        )
        self.simulation_canvas.stroke_rects(
            x=self.tiles_x,
            y=self.tiles_y,
            width=self.size,
            height=self.size,
        )

        # get arrow directions for tiles
        arg_max_q = [
            tile["q_values"].argmax() for idx, tile in enumerate(self.grid) if idx < 37
        ]
        arrows = [arrow[arg] for arg, arrow in zip(arg_max_q, self.arrows)]

        # draw arrows
        self.simulation_canvas.fill_style = self.arrow_color
        self.simulation_canvas.fill_polygons(arrows)

        # write q_values
        for idx, tile in enumerate(self.grid):
            if idx > 36:
                break
            # q move top
            self.simulation_canvas.fill_text(
                x=tile["x"] + int(self.size / 2),
                y=tile["y"] + self.font_size,
                text=f'{tile["q_values"][0]:.{self.show_decimals}f}',
            )
            # q move right
            self.simulation_canvas.fill_text(
                x=tile["x"] + self.size - self.font_size * self.show_decimals * 0.75,
                y=tile["y"] + int(self.size / 2),
                text=f'{tile["q_values"][1]:.{self.show_decimals}f}',
            )
            # q move down
            self.simulation_canvas.fill_text(
                x=tile["x"] + int(self.size / 2),
                y=tile["y"] + self.size - self.font_size,
                text=f'{tile["q_values"][2]:.{self.show_decimals}f}',
            )
            # q move left
            self.simulation_canvas.fill_text(
                x=tile["x"] + self.font_size * self.show_decimals * 0.75,
                y=tile["y"] + int(self.size / 2),
                text=f'{tile["q_values"][3]:.{self.show_decimals}f}',
            )
